# Copyright 2017 Kakao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import io
import logging
import os
import platform
import subprocess

from setuptools import setup, Extension

from Cython.Build import cythonize

NAME = 'n2'
VERSION = '0.1.4'

def long_description():
    with io.open('README.rst', 'r', encoding='utf-8') as f:
        readme = f.read()
    return readme

def define_extensions(**kwargs):
    libraries = []
    extra_link_args = []
    extra_compile_args = ['-std=c++11', '-O3', '-fPIC', '-march=native']
    extra_link_args.append('-fopenmp')
    extra_compile_args.append('-fopenmp')

    sources = ['./src/base.cc', './src/distance.cc', './src/heuristic.cc',
        './src/hnsw.cc', './src/hnsw_node.cc',  './src/mmap.cc',
        './bindings/python/n2.pyx']

    client_ext = Extension(name='n2',
                           sources=sources,
                           extra_compile_args=extra_compile_args,
                           libraries=libraries,
                           extra_link_args=extra_link_args,
                           include_dirs=['./include/', './third_party/spdlog/include/'],
                           language="c++",)
    return cythonize(client_ext)

setup(
    name=NAME,
    version=VERSION,
    description='Approximate Nearest Neighbor library',
    long_description=long_description(),
    author='Kakao.corp',
    author_email='recotech.kakao@gmail.com',
    license='Apache License 2.0',
    install_requires=[
        "cython",
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Cython',
        'Topic :: Software Development :: Libraries :: Python Modules'],

    keywords='Approximate Nearest Neighbor',
    ext_modules=define_extensions(),
)



# -*- coding: utf-8 -*-
import os
import abc
import psutil

import h5py
import numpy as np

from buffalo.data import prepro
from buffalo.misc import aux, log
from buffalo.data.fileio import chunking_into_bins, sort_and_compressed_binarization


class Data(object):
    def __init__(self, opt, *args, **kwargs):
        self.opt = aux.Option(opt)
        self.tmp_root = opt.data.tmp_dir
        if not os.path.isdir(self.tmp_root):
            os.makedirs(self.tmp_root)
        self.handle = None
        self.header = None
        self.prepro = prepro.PreProcess(self.opt.data)
        self.value_prepro = self.prepro
        if self.opt.data.value_prepro:
            self.prepro = getattr(prepro, self.opt.data.value_prepro.name)(self.opt.data.value_prepro)
            self.value_prepro = self.prepro
        self.data_type = None

    @abc.abstractmethod
    def create_database(self, filename, **kwargs):
        pass

    def show_info(self):
        header = self.get_header()
        vali_size = 0
        if self.has_group('vali'):
            g = self.get_group('vali')
            vali_size = g.attrs['num_samples']
        info = '{name} Header({users}, {items}, {nnz}) Validation({vali} samples)'
        info = info.format(name=self.name,
                           users=header['num_users'],
                           items=header['num_items'],
                           nnz=header['num_nnz'],
                           vali=vali_size)
        return info

    def open(self, data_path):
        self.handle = h5py.File(data_path, 'r')
        self.path = data_path
        self.verify()

    def verify(self):
        assert self.handle, 'Database is not opened'
        if self.get_header()['completed'] != 1:
            raise RuntimeError('Database is corrupted or partially built. Please try again, after remove it.')

    def get_header(self):
        assert self.handle, 'Database is not opened'
        if not self.header:
            self.header = {'num_nnz': self.handle.attrs['num_nnz'],
                           'num_users': self.handle.attrs['num_users'],
                           'num_items': self.handle.attrs['num_items'],
                           'completed': self.handle.attrs['completed']}
        return self.header

    def get_scale_info(self, with_sppmi=False, chunk_size=100000):
        ret = {k: self.handle.attrs[k] for k in ["num_users", "num_items", "num_nnz", "sppmi_nnz"]}
        if with_sppmi:
            ret["sppmi_nnz"] = self.handle.attrs["sppmi_nnz"]
        db = self.handle["rowwise"]
        num_nnz = ret["num_nnz"]
        vsum = 0.0
        for offset in range(0, num_nnz, chunk_size):
            limit = min(num_nnz, offset + chunk_size)
            vsum += np.sum(db["val"][offset: limit])
        ret["vsum"] = vsum
        return ret

    def get_group(self, group_name='rowwise'):
        assert group_name in ['rowwise', 'colwise', 'vali', 'idmap', 'sppmi'], 'Unexpected group_name: {}'.format(group_name)
        assert self.handle, 'DB is not opened'
        group = self.handle[group_name]
        return group

    def has_group(self, name):
        return name in self.handle

    def _iterate_matrix(self, axis, use_repr_name, userids, itemids):
        assert axis in ['rowwise', 'colwise'], 'Unexpected data axis: {}'.format(axis)
        assert self.handle, 'Database is not opened'
        group = self.handle[axis]
        data_index = 0
        for u, end in enumerate(group['indptr']):
            keys = group['key'][data_index:end]
            vals = group['val'][data_index:end]
            if use_repr_name:
                for k, v in zip(keys, vals):
                    yield userids(u), itemids(k), v
            else:
                for k, v in zip(keys, vals):
                    yield u, k, v
            data_index = end

    def _iterate_stream(self, axis, use_repr_name, userids, itemids):
        assert axis in ['rowwise'], 'Unexpected date axis: {}'.format(axis)
        assert self.handle, 'Database is not opened'
        group = self.handle[axis]
        data_index = 0
        for u, end in enumerate(group['indptr']):
            keys = group['key'][data_index:end]
            if use_repr_name:
                for k in keys:
                    yield userids(u), itemids(k)
            else:
                for k in keys:
                    yield u, k
            data_index = end

    def iterate(self, axis='rowwise', use_repr_name=False) -> [int, int, float]:
        """Iterate over datas

        args:
            group: which data group
            use_repr_name: return representive name of internal ids
        """
        userids, itemids = None, None
        idmap = self.get_group('idmap')
        if use_repr_name:
            userids = lambda x: str(x)
            if idmap['rows'].shape[0] != 0:
                userids = lambda x: idmap['rows'][x].decode('utf-8', 'ignore')
            itemids = lambda x: str(x)
            if idmap['cols'].shape[0] != 0:
                itemids = lambda x: idmap['cols'][x].decode('utf-8', 'ignore')
            if axis == 'colwise':
                userids, itemids = itemids, userids

        if self.opt.data.internal_data_type == 'matrix':
            return self._iterate_matrix(axis, use_repr_name, userids, itemids)
        elif self.opt.data.internal_data_type == 'stream':
            return self._iterate_stream(axis, use_repr_name, userids, itemids)

    def get(self, index, axis='rowwise') -> [int, int, float]:
        if self.opt.data.internal_data_type == 'matrix':
            assert axis in ['rowwise', 'colwise'], 'Unexpected data axis: {}'.format(axis)
            assert self.handle, 'Database is not opened'
            group = self.handle[axis]
            begin = 0 if index == 0 else group['indptr'][index - 1]
            end = group['indptr'][index]
            keys = group['key'][begin:end]
            vals = group['val'][begin:end]
            return (keys, vals)
        elif self.opt.data.internal_data_type == 'stream':
            assert axis in ['rowwise'], 'Unexpected date axis: {}'.format(axis)
            assert self.handle, 'Database is not opened'
            group = self.handle[axis]
            begin = 0 if index == 0 else group['indptr'][index - 1]
            end = group['indptr'][index]
            keys = group['key'][begin:end]
            return (keys,)

    def close(self):
        if self.handle:
            self.handle = None
            self.header = None

    def _create_database(self, path, **kwargs):
        # Create database structure
        f = h5py.File(path, 'w')
        self.path = path
        num_users, num_items, num_nnz = kwargs['num_users'], kwargs['num_items'], kwargs['num_nnz']

        f.create_group('rowwise')
        f.create_group('colwise')
        chunk_size = (min(1024 ** 3, num_nnz),)
        for g in [f['rowwise'], f['colwise']]:
            g.create_dataset('key', (num_nnz,), dtype='int32', maxshape=(num_nnz,), chunks=chunk_size)
            g.create_dataset('val', (num_nnz,), dtype='float32', maxshape=(num_nnz,), chunks=chunk_size)
        f['rowwise'].create_dataset('indptr', (num_users,), dtype='int64', maxshape=(num_users,))
        f['colwise'].create_dataset('indptr', (num_items,), dtype='int64', maxshape=(num_items,))

        num_nnz = self._create_validation(f, **kwargs)

        f.attrs['num_users'] = num_users
        f.attrs['num_items'] = num_items
        f.attrs['num_nnz'] = num_nnz
        f.attrs['completed'] = 0

        iid_max_col = kwargs['iid_max_col']
        uid_max_col = kwargs['uid_max_col']
        idmap = f.create_group('idmap')
        idmap.create_dataset('rows', (num_users,), dtype='S%s' % uid_max_col,
                             maxshape=(num_users,))
        idmap.create_dataset('cols', (num_items,), dtype='S%s' % iid_max_col,
                             maxshape=(num_items,))
        return f

    def _create_validation(self, f, **kwargs):
        if not self.opt.data.validation:
            return kwargs['num_nnz']
        num_nnz = kwargs['num_nnz']
        method = self.opt.data.validation.name

        f.create_group('vali')
        g = f['vali']
        g.attrs['method'] = method
        g.attrs['n'] = 0
        if method == 'sample':
            sz = min(self.opt.data.validation.max_samples,
                     int(num_nnz * self.opt.data.validation.p))
            g.create_dataset('row', (sz,), dtype='int32', maxshape=(sz,))
            g.create_dataset('col', (sz,), dtype='int32', maxshape=(sz,))
            g.create_dataset('val', (sz,), dtype='float32', maxshape=(sz,))
            g.create_dataset('indexes', (sz,), dtype='int64', maxshape=(sz,))
            # Watch out, create_working_data cannot deal with last line of data
            # for validation data thus we have to reduce index 1.
            g['indexes'][:] = np.random.choice(num_nnz - 1, sz, replace=False)
            num_nnz -= sz
        elif method in ['newest']:
            sz = kwargs['num_validation_samples']
            g.create_dataset('row', (sz,), dtype='int32', maxshape=(sz,))
            g.create_dataset('col', (sz,), dtype='int32', maxshape=(sz,))
            g.create_dataset('val', (sz,), dtype='float32', maxshape=(sz,))
            g.attrs['n'] = self.opt.data.validation.n
            # We don't need to reduce sample size for validation samples. It
            # already applied on caller side.
        g.attrs['num_samples'] = sz
        return num_nnz

    def fill_validation_data(self, db, validation_data):
        if not validation_data:
            return
        validation_data = [line.strip().split() for line in validation_data]
        assert len(validation_data) == db['vali'].attrs['num_samples'], 'Mimatched validation data'
        db['vali']['row'][:] = [int(r) - 1 for r, _, _ in validation_data]  # 0-based
        db['vali']['col'][:] = [int(c) - 1 for _, c, _ in validation_data]  # 0-based
        V = np.array([float(v) for _, _, v in validation_data], dtype=np.float32)
        db['vali']['val'][:] = self.value_prepro(V)

    def _sort_and_compressed_binarization(self, mm_path, num_lines, max_key, sort_key):
        num_workers = psutil.cpu_count()
        merged_bin = sort_and_compressed_binarization(
            mm_path,
            self.tmp_root,
            num_lines, max_key, sort_key, num_workers)
        return merged_bin

    def _load_compressed_triplet_bin(self, db, bin_path, num_lines, max_key, is_colwise=0):
        self.logger.info('Load triplet bin: %s' % bin_path)
        INDPTR_SIZE = 8
        RECORD_SIZE = 8
        with open(bin_path, 'rb') as fin:
            estimated_size = num_lines * 8 + max_key * 8
            total_size = fin.seek(0, 2)
            fin.seek(0, 0)
            assert estimated_size == total_size, f'Not valid file size {total_size} (excepted: {estimated_size})'
            indptr = np.frombuffer(fin.read(max_key * 8),
                                   dtype=np.int64,
                                   count=max_key)
            data = np.frombuffer(fin.read(),
                                 dtype=np.dtype([('i', 'i'),
                                                 ('v', 'f')]),
                                 count=num_lines).copy()
            I, V = data['i'], data['v']
            I -= 1
            V = self.value_prepro(V)
            db['key'][:num_lines] = I
            db['val'][:num_lines] = V
            db['indptr'][:max_key] = indptr
        os.remove(bin_path)

    def _chunking_into_bins(self, mm_path, num_lines, max_key, sep_idx):
        num_workers = psutil.cpu_count()
        while num_workers > 20:
            num_workers = int(num_workers / 2)
        num_chunks = num_workers * 2
        self.logger.info(f'Dividing into {num_chunks} chunks...')
        job_files = chunking_into_bins(mm_path,
                                       self.tmp_root,
                                       total_lines=num_lines,
                                       num_chunks=num_chunks,
                                       sep_idx=sep_idx,
                                       num_workers=num_workers)
        return job_files

    def _build_compressed_triplets(self, db, job_files, num_lines, max_key, is_colwise=0):
        self.logger.info('Total job files: %s' % len(job_files))
        with log.pbar(log.INFO, total=len(job_files), mininterval=10) as pbar:
            indptr_index = 0
            data_index = 0
            RECORD_SIZE = 12
            prev_key = 0
            for job in job_files:
                with open(job, 'rb') as fin:
                    total_size = fin.seek(0, 2)
                    if total_size == 0:
                        continue
                    total_records = int(total_size / RECORD_SIZE)
                    fin.seek(0, 0)
                    data = np.frombuffer(fin.read(),
                                         dtype=np.dtype([('u', 'i'),
                                                         ('i', 'i'),
                                                         ('v', 'f')]),
                                         count=total_records).copy()
                    U, I, V = data['u'], data['i'], data['v']
                    if is_colwise:
                        U, I = I, U
                    U -= 1
                    I -= 1
                    V = self.value_prepro(V)
                    self.logger.debug("minU: {}, maxU: {}".format(U[0], U[-1]))
                    assert data_index + total_records <= num_lines, 'Requests data size(%s) exceed capacity(%s)' % (data_index + total_records, num_lines)
                    db['key'][data_index:data_index + total_records] = I
                    db['val'][data_index:data_index + total_records] = V
                    indptr = [data_index for j in range(U[0] - prev_key)]
                    indptr += [data_index + i
                               for i in range(1, total_records)
                               for j in range(U[i] - U[i-1])]
                    db['indptr'][indptr_index:indptr_index + len(indptr)] = indptr
                    assert indptr_index + len(indptr) <= max_key
                    data_index += total_records
                    indptr_index += len(indptr)
                    prev_key = U[-1]
                pbar.update(1)
            db["indptr"][indptr_index:] = data_index
        for path in job_files:
            os.remove(path)

    def _build_data(self,
                    db,
                    working_data_path,
                    validation_data,
                    target_groups=['rowwise', 'colwise'],
                    sort=True):
        available_mb = psutil.virtual_memory().available / 1024 / 1024
        approximated_data_mb = 0
        with open(working_data_path, 'rb') as fin:
            fin.seek(0, 2)
            approximated_data_mb = db.attrs['num_nnz'] * 3 / 1024 / 1024
        buffer_mb = int(max(1024, available_mb * 0.75))
        # for each sides
        for group, sep_idx, max_key in [('rowwise', 0, db.attrs['num_users']),
                                        ('colwise', 1, db.attrs['num_items'])]:
            if group not in target_groups:
                continue
            self.logger.info(f'Building compressed triplets for {group}...')
            self.logger.info('Preprocessing...')
            self.prepro.pre(db)
            if approximated_data_mb * 1.2 < available_mb:
                self.logger.info('In-memory Compressing ...')
                merged_bin = self._sort_and_compressed_binarization(
                    working_data_path,
                    db.attrs['num_nnz'],
                    max_key,
                    sort_key=sep_idx + 1 if sort else -1)
                self._load_compressed_triplet_bin(
                    db[group], merged_bin,
                    num_lines=db.attrs['num_nnz'],
                    max_key=max_key,
                    is_colwise=sep_idx)
            else:
                self.logger.info('Disk-based Compressing...')
                if sort:
                    aux.psort(working_data_path,
                              tmp_dir=self.opt.data.tmp_dir,
                              key=sep_idx + 1,
                              buffer_mb=buffer_mb)
                job_files = self._chunking_into_bins(working_data_path,
                                                     db.attrs['num_nnz'],
                                                     max_key,
                                                     sep_idx=sep_idx)
                self._build_compressed_triplets(db[group],
                                                job_files,
                                                num_lines=db.attrs['num_nnz'],
                                                max_key=max_key,
                                                is_colwise=sep_idx)
            self.prepro.post(db[group])
            if group == 'rowwise':
                self.fill_validation_data(db, validation_data)
            self.logger.info('Finished')


class DataOption(object):
    def is_valid_option(self, opt) -> bool:
        """General type/logic checking"""

        if 'validation' in opt['data']:
            assert opt['data']['validation']['name'] in ['sample', 'newest'], 'Unknown validation.name.'
            if opt['data']['validation']['name'] == 'sample':
                assert hasattr(opt['data']['validation'], 'max_samples'), 'max_samples not defined on data.validation.'
                assert isinstance(opt['data']['validation']['max_samples'], int), 'invalid type for data.validation.max_samples'
                assert hasattr(opt['data']['validation'], 'p'), 'not defined on data.validation.p'
                assert isinstance(opt['data']['validation']['p'], float), 'invalid type for data.validation.p'
            if opt['data']['validation']['name'] in ['newest']:
                assert hasattr(opt['data']['validation'], 'max_samples'), 'max_samples not defined on data.validation.'
                assert isinstance(opt['data']['validation']['max_samples'], int), 'invalid type for data.validation.max_samples'
                assert hasattr(opt['data']['validation'], 'n'), 'not defined on data.validation.n'
                assert isinstance(opt['data']['validation']['n'], int), 'invalid type for data.validation.n'
        return True



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
#-*- coding: utf-8 -*-
from libc.stdint cimport int64_t
from libcpp.vector cimport vector
from libcpp.string cimport string


cdef extern from "fileio.hpp" namespace "fileio":
    vector[string] _chunking_into_bins(string, string, int64_t, int, int, int)
    string _sort_and_compressed_binarization(string, string, int64_t, int, int, int)


def chunking_into_bins(path, to_dir, total_lines, num_chunks, sep_idx, num_workers):
    assert num_chunks % num_workers == 0, 'The number of chunks must be a multiple of the number of workers'
    return _chunking_into_bins(bytes(path, 'utf-8'),
                               bytes(to_dir, 'utf-8'),
                               total_lines,
                               num_chunks,
                               sep_idx,
                               num_workers)


def sort_and_compressed_binarization(path, to_dir, total_lines, max_key, sort_key, num_workers):
    return _sort_and_compressed_binarization(
        bytes(path, 'utf-8'),
        bytes(to_dir, 'utf-8'),
        total_lines,
        max_key,
        sort_key,
        num_workers)



import gzip
import numpy
import time
import os
import multiprocessing
import argparse
import pickle
import resource
import random
import math
import logging
import shutil
import subprocess
import sys
import tarfile

from contextlib import closing

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

try:
    xrange
except NameError:
    xrange = range

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

# Set resource limits to prevent memory bombs
memory_limit = 12 * 2**30
soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
if soft == resource.RLIM_INFINITY or soft >= memory_limit:
    logging.debug('resetting memory limit from {0} to {1}. '.format(
                  soft, memory_limit))
    resource.setrlimit(resource.RLIMIT_DATA, (memory_limit, hard))

INDEX_DIR = 'indices'
DATA_DIR = './datasets/'
YOUTUBE_DIR = DATA_DIR + 'youtube.txt'


class BaseANN(object):
    def use_threads(self):
        return True


class BruteForceBLAS(BaseANN):
    """kNN search that uses a linear scan = brute force."""

    def __init__(self, metric, precision=numpy.float32):
        if metric not in ('angular', 'euclidean'):
            raise NotImplementedError(
                "BruteForceBLAS doesn't support metric %s" %
                metric)
        self._metric = metric
        self._precision = precision
        self.name = 'BruteForceBLAS()'

    def fit(self, X):
        """Initialize the search index."""
        lens = (X ** 2).sum(-1)  # precompute (squared) length of each vector
        if self._metric == 'angular':
            # normalize index vectors to unit length
            X /= numpy.sqrt(lens)[..., numpy.newaxis]
            self.index = numpy.ascontiguousarray(X, dtype=self._precision)
        elif self._metric == 'euclidean':
            self.index = numpy.ascontiguousarray(X, dtype=self._precision)
            self.lengths = numpy.ascontiguousarray(lens, dtype=self._precision)
        else:
            assert False, "invalid metric"  # shouldn't get past the constructor!

    def query(self, v, n):
        """Find indices of `n` most similar vectors from the index to query vector `v`."""
        v = numpy.ascontiguousarray(v, dtype=self._precision)
        # use same precision for query as for index
        # HACK we ignore query length as that's a constant not affecting the
        # final ordering
        if self._metric == 'angular':
            # argmax_a cossim(a, b) = argmax_a dot(a, b) / |a||b| = argmin_a -dot(a, b)
            dists = -numpy.dot(self.index, v)
        elif self._metric == 'euclidean':
            # argmin_a (a - b)^2 = argmin_a a^2 - 2ab + b^2 = argmin_a a^2 - 2ab
            dists = self.lengths - 2 * numpy.dot(self.index, v)
        else:
            assert False, "invalid metric"  # shouldn't get past the constructor!
        # partition-sort by distance, get `n` closest
        indices = numpy.argpartition(dists, n)[:n]
        # sort `n` closest into correct order
        return sorted(indices, key=lambda index: dists[index])


class N2(BaseANN):
    def __init__(self, m, ef_construction, n_threads, ef_search, metric):
        self._m = m
        self._m0 = m * 2
        self._ef_construction = ef_construction
        self._n_threads = n_threads
        self._ef_search = ef_search
        self._index_name = os.path.join(
            INDEX_DIR, "youtube_n2_M%d_efCon%d_n_thread%s" %
            (m, ef_construction, n_threads))
        self.name = "N2_M%d_efCon%d_n_thread%s_efSearch%d" % (
            m, ef_construction, n_threads, ef_search)
        self._metric = metric

        d = os.path.dirname(self._index_name)
        if not os.path.exists(d):
            os.makedirs(d)

    def fit(self, X):
        from n2 import HnswIndex
        if self._metric == 'euclidean':
            self._n2 = HnswIndex(X.shape[1], 'L2')
        else:
            self._n2 = HnswIndex(X.shape[1])
        if os.path.exists(self._index_name):
            logging.debug("Loading index from file")
            self._n2.load(self._index_name)
        else:
            logging.debug(
                "Index file is not exist: {0}".format(
                    self._index_name))
            logging.debug("Start fitting")

            for i, x in enumerate(X):
                self._n2.add_data(x.tolist())
            self._n2.build(
                m=self._m,
                max_m0=self._m0,
                ef_construction=self._ef_construction,
                n_threads=self._n_threads)
            self._n2.save(self._index_name)

    def query(self, v, n):
        return self._n2.search_by_vector(v.tolist(), n, self._ef_search)

    def __str__(self):
        return self.name


class NmslibReuseIndex(BaseANN):
    def __init__( self, metric, method_name, index_param, save_index,query_param):
        self._nmslib_metric = {
            'angular': 'cosinesimil',
            'euclidean': 'l2'}[metric]
        self._method_name = method_name
        self._save_index = save_index
        self._index_param = index_param
        self._query_param = query_param
        self.name = 'Nmslib(method_name=%s, index_param=%s, query_param=%s)' % (
            method_name, index_param, query_param)
        self._index_name = os.path.join(
            INDEX_DIR, "youtube_nmslib_%s_%s_%s" %
            (self._method_name, metric, '_'.join(
                self._index_param)))

        d = os.path.dirname(self._index_name)
        if not os.path.exists(d):
            os.makedirs(d)

    def fit(self, X):
        import nmslib
        self._index = nmslib.init(
            self._nmslib_metric,
            [],
            self._method_name,
            nmslib.DataType.DENSE_VECTOR,
            nmslib.DistType.FLOAT)

        for i, x in enumerate(X):
            nmslib.addDataPoint(self._index, i, x.tolist())

        if os.path.exists(self._index_name):
            logging.debug("Loading index from file")
            nmslib.loadIndex(self._index, self._index_name)
        else:
            logging.debug("Create Index")
            nmslib.createIndex(self._index, self._index_param)
            if self._save_index:
                nmslib.saveIndex(self._index, self._index_name)

        nmslib.setQueryTimeParams(self._index, self._query_param)

    def query(self, v, n):
        import nmslib
        return nmslib.knnQuery(self._index, n, v.tolist())

    def freeIndex(self):
        import nmslib
        nmslib.freeIndex(self._index)


class Annoy(BaseANN):
    def __init__(self, metric, n_trees, search_k):
        self._n_trees = n_trees
        self._search_k = search_k
        self._metric = metric
        self._index_name = os.path.join(
            INDEX_DIR, "youtube_annoy_%s_tree%d" %
            (metric, n_trees))
        self.name = 'Annoy(n_trees=%d, search_k=%d)' % (n_trees, search_k)

        d = os.path.dirname(self._index_name)
        if not os.path.exists(d):
            os.makedirs(d)

    def fit(self, X):
        import annoy
        self._annoy = annoy.AnnoyIndex(f=X.shape[1], metric=self._metric)
        if os.path.exists(self._index_name):
            logging.debug("Loading index from file")
            self._annoy.load(self._index_name)
        else:
            logging.debug("Index file not exist start fitting!!")
            for i, x in enumerate(X):
                self._annoy.add_item(i, x.tolist())
            self._annoy.build(self._n_trees)
            self._annoy.save(self._index_name)

    def query(self, v, n):
        return self._annoy.get_nns_by_vector(v.tolist(), n, self._search_k)


def get_fn(base, args):
    fn = os.path.join(base, args.dataset)

    if args.limit != -1:
        fn += '-%d' % args.limit
    if os.path.exists(fn + '.gz'):
        fn += '.gz'
    else:
        fn += '.txt'

    d = os.path.dirname(fn)
    if not os.path.exists(d):
        os.makedirs(d)

    return fn


def run_algo(args, library, algo, results_fn):
    pool = multiprocessing.Pool()
    X_train, X_test = get_dataset(args.dataset, args.limit)
    pool.close()
    pool.join()

    t0 = time.time()
    algo.fit(X_train)
    build_time = time.time() - t0
    logging.debug('Built index in {0}'.format(build_time))
    best_search_time = float('inf')
    best_precision = 0.0  # should be deterministic but paranoid
    try_count = 3
    # Do multiple times to warm up page cache, use fastest
    for i in xrange(try_count):
        results = []
        search_time = 0.0
        current_query = 1
        total_queries = len(queries)
        for j in range(total_queries):
            v, correct = queries[j]
            sys.stdout.write(
                "Querying: %d / %d \r" %
                (current_query, total_queries))
            t0 = time.time()
            found = algo.query(v, GT_SIZE)
            search_time += (time.time() - t0)
            if len(found) < len(correct):
                logging.debug(
                    'found: {0}, correct: {1}'.format(
                        len(found), len(correct)))
            current_query += 1
            results.append(len(set(found).intersection(correct)))

        k = float(sum(results))
        search_time /= len(queries)
        precision = k / (len(queries) * GT_SIZE)
        best_search_time = min(best_search_time, search_time)
        best_precision = max(best_precision, precision)
        sys.stdout.write(
            '*[%d/%d][algo: %s] search time: %s, precision: %.5f \r' %
            (i + 1, try_count, str(algo), str(search_time), precision))
        sys.stdout.write('\n')
    output = [library, algo.name, build_time, best_search_time, best_precision]
    logging.debug(str(output))
    f = open(results_fn, 'a')
    f.write('\t'.join(map(str, output)) + '\n')
    f.close()
    logging.debug('Summary: {0}'.format('\t'.join(map(str, output))))


def get_dataset(which='glove', limit=-1, random_state=3, test_size=10000):
    cache = 'queries/%s-%d-%d-%d.npz' % (which, test_size, limit, random_state)
    if os.path.exists(cache):
        v = numpy.load(cache)
        X_train = v['train']
        X_test = v['test']
        logging.debug('{0} {1}'.format(X_train.shape, X_test.shape))
        return X_train, X_test
    local_fn = os.path.join('datasets', which)
    if os.path.exists(local_fn + '.gz'):
        f = gzip.open(local_fn + '.gz')
    else:
        f = open(local_fn + '.txt')

    X = []
    for i, line in enumerate(f):
        v = [float(x) for x in line.strip().split()]
        X.append(v)
        if limit != -1 and len(X) == limit:
            break

    X = numpy.vstack(X)
    import sklearn.cross_validation

    # Here Erik is most welcome to use any other random_state
    # However, it is best to use a new random seed for each major re-evaluation,
    # so that we test on a trully bind data.
    X_train, X_test = sklearn.cross_validation.train_test_split(
        X, test_size=test_size, random_state=random_state)
    X_train = X_train.astype(numpy.float)
    X_test = X_test.astype(numpy.float)
    numpy.savez(cache, train=X_train, test=X_test)
    return X_train, X_test


def get_queries(args):
    logging.debug('computing queries with correct results...')

    bf = BruteForceBLAS(args.distance)
    X_train, X_test = get_dataset(which=args.dataset, limit=args.limit)

    # Prepare queries
    bf.fit(X_train)
    queries = []
    total_queries = len(X_test)
    for x in X_test:
        correct = bf.query(x, GT_SIZE)
        queries.append((x, correct))
        sys.stdout.write(
            'computing queries %d/%d ...\r' %
            (len(queries), total_queries))

    sys.stdout.write('\n')
    return queries


def get_fn(base, args):
    fn = os.path.join(base, args.dataset)

    if args.limit != -1:
        fn += '-%d' % args.limit
    if os.path.exists(fn + '.gz'):
        fn += '.gz'
    else:
        fn += '.txt'

    d = os.path.dirname(fn)
    if not os.path.exists(d):
        os.makedirs(d)

    return fn


if __name__ == '__main__':
    global GT_SIZE
    parser = argparse.ArgumentParser()
    parser.add_argument('--algo', help='Algorithm', type=str)
    parser.add_argument('--limit', help='Limit', type=int, default=-1)
    parser.add_argument('--n_threads', help='Number of threads', type=int,default=20)
    args = parser.parse_args()
    args.dataset = 'youtube'
    args.distance = 'angular'
    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR)

    GT_SIZE = 100

    if not os.path.exists(YOUTUBE_DIR):
        raise IOError(
            'Please follow the instructions in the guide to download the YouTube dataset.')

    results_fn = get_fn('results', args)
    queries_fn = get_fn('queries', args)
    logging.debug(
        'storing queries in {0} and results in {1}.'.format(
            queries_fn, results_fn))

    if not os.path.exists(queries_fn):
        queries = get_queries(args)
        with open(queries_fn, 'wb') as f:
            pickle.dump(queries, f)
    else:
        queries = pickle.load(open(queries_fn, 'rb'))

    logging.debug('got {0} queries'.format(len(queries)))

    algos = {
        'annoy': [ Annoy('angular', n_trees, search_k)
                  for n_trees in [10]
                  for search_k in [ 7, 3000, 50000, 200000, 500000]
               ], 
        'n2': [ N2(M, ef_con, args.n_threads, ef_search, 'angular') 
               for M, ef_con in [ (12, 100)] 
               for ef_search in [10, 100, 1000, 10000, 100000]
            ], 
        'nmslib': []}

    MsPostsEfs = [
        ({'M': 12,
          'post': 0,
          'indexThreadQty': args.n_threads,
          'delaunay_type': 2,
          'efConstruction': 100,
          },
         [10, 100, 1000, 10000, 100000],
         ),
    ]

    for oneCase in MsPostsEfs:
        for ef in oneCase[1]:
            params = ['%s=%s' % (k, str(v)) for k, v in oneCase[0].items()]
            algos['nmslib'].append(
                NmslibReuseIndex( 'angular', 'hnsw', params, True, ['ef=%d' % ef]))

    algos_flat = []

    if args.algo:
        print('running only: %s' % str(args.algo))
        algos = {args.algo: algos[args.algo]}

    for library in algos.keys():
        for algo in algos[library]:
            algos_flat.append((library, algo))

    random.shuffle(algos_flat)
    logging.debug('order: %s' % str([a.name for l, a in algos_flat]))

    for library, algo in algos_flat:
        logging.debug(algo.name)
        # Spawn a subprocess to force the memory to be reclaimed at the end
        p = multiprocessing.Process(
            target=run_algo, args=(
                args, library, algo, results_fn))
        p.start()
        p.join()



# -*- coding: utf-8 -*-
import numpy as np


class PreProcess(object):
    def __init__(self, opt):
        self.opt = opt

    def pre(self, header):
        pass

    def __call__(self, v):
        return v

    def post(self, db):
        pass


class OneBased(PreProcess):
    def __init__(self, opt):
        super(OneBased, self).__init__(opt)

    def pre(self, header):
        return

    def __call__(self, v):
        v[:] = 1.0
        return v

    def post(self, db):
        return


class MinMaxScalar(PreProcess):
    def __init__(self, opt):
        super(MinMaxScalar, self).__init__(opt)
        self.value_min = 987654321.0
        self.value_max = 0.0

    def pre(self, header):
        return

    def __call__(self, V):
        self.value_min = min(self.value_min, np.min(V))
        self.value_max = max(self.value_max, np.max(V))
        return V

    def post(self, db):
        if self.value_max - self.value_min < 0.00000001:
            db['val'][:] = 1.0
            db['val'][:] *= self.opt.max
            return
        sz = db['val'].shape[0]
        per = db['val'].chunks[0]
        for idx in range(0, sz, per):
            chunk = db['val'][idx:idx + per]
            chunk = (chunk - self.value_min) / (self.value_max - self.value_min)
            chunk = chunk * (self.opt.max - self.opt.min) + self.opt.min
            db['val'][idx:idx + per] = chunk


class ImplicitALS(PreProcess):
    def __init__(self, opt):
        super(ImplicitALS, self).__init__(opt)

    def pre(self, header):
        return

    def __call__(self, V):
        return np.log(1 + V / self.opt.epsilon)

    def post(self, db):
        return


class SPPMI(PreProcess):
    """NotImplemented Yet"""
    def __init__(self, opt):
        super(SPPMI, self).__init__(opt)

    def pre(self, header):
        return

    def __call__(self, v):
        return v

    def post(self, db):
        return



# -*- coding: utf-8 -*-
import time
import json
import logging

import tqdm
import numpy as np
from hyperopt import STATUS_OK as HOPT_STATUS_OK

import buffalo.data
from buffalo.misc import aux, log
from buffalo.data.base import Data
from buffalo.algo._cfr import CyCFR
from buffalo.evaluate import Evaluable
from buffalo.algo.options import CFROption
from buffalo.algo.optimize import Optimizable
from buffalo.data.buffered_data import BufferedDataMatrix
from buffalo.algo.base import Algo, Serializable, TensorboardExtention


class CFR(Algo, CFROption, Evaluable, Serializable, Optimizable, TensorboardExtention):
    """Python implementation for CoFactor.

    Reference: Factorization Meets the Item Embedding:
        Regularizing Matrix Factorization with Item Co-occurrence

    Paper link: http://dawenl.github.io/publications/LiangACB16-cofactor.pdf
    """
    def __init__(self, opt_path=None, *args, **kwargs):
        Algo.__init__(self, *args, **kwargs)
        CFROption.__init__(self, *args, **kwargs)
        Evaluable.__init__(self, *args, **kwargs)
        Serializable.__init__(self, *args, **kwargs)
        Optimizable.__init__(self, *args, **kwargs)
        if opt_path is None:
            opt_path = CFROption().get_default_option()

        self.logger = log.get_logger('CFR')

        # put options into cython class with type assertion
        # see comments on options.py for the description of each parameter
        self.opt, self.opt_path = self.get_option(opt_path)
        self.obj = CyCFR()
        # check the validity of option
        self.is_valid_option(self.opt)
        assert self.obj.init(self.opt_path.encode("utf8")), "putting parameter to cython object failed"

        # ensure embedding matrix is initialzed for preventing segmentation fault
        self.is_initialized = False

        self.data = None
        data = kwargs.get('data')
        data_opt = self.opt.get('data_opt')
        data_opt = kwargs.get('data_opt', data_opt)
        if data_opt:
            self.data = buffalo.data.load(data_opt)
            self.data.create()
        elif isinstance(data, Data):
            self.data = data
        self.logger.info('CFR ({})'.format(json.dumps(self.opt, indent=2)))
        if self.data:
            self.logger.info(self.data.show_info())
            assert self.data.data_type in ['stream']

    @staticmethod
    def new(path, data_fields=[]):
        return CFR.instantiate(CFROption, path, data_fields)

    def set_data(self, data):
        assert isinstance(data, aux.data.Data), 'Wrong instance: {}'.format(type(data))
        self.data = data

    def normalize(self, group='item'):
        assert group in ["user", "item", "context"], \
            f"group ({group}) is not properly provided"
        if group == 'user' and not self.opt._nrz_U:
            self.U = self._normalize(self.U)
            self.opt._nrz_U = True
        elif group == 'item' and not self.opt._nrz_I:
            self.I = self._normalize(self.I)
            self.opt._nrz_I = True
        elif group == 'context' and not self.opt._nrz_C:
            self.C = self._normalize(self.C)
            self.opt._nrz_C = True

    def initialize(self):
        super().initialize()
        assert self.data, 'Data is not setted'
        header = self.data.get_header()
        num_users, num_items, d = \
            header["num_users"], header["num_items"], self.opt.d
        for attr, shape, name in [('U', (num_users, d), "user"),
                                  ('I', (num_items, d), "item"),
                                  ('C', (num_items, d), "context"),
                                  ('Ib', (num_items, 1), "item_bias"),
                                  ('Cb', (num_items, 1), "context_bias")]:
            setattr(self, attr, None)
            F = np.random.normal(scale=1.0/(d ** 2), size=shape).astype(np.float32)
            setattr(self, attr, F)
            self.obj.set_embedding(getattr(self, attr), name.encode("utf8"))
        self.P = self.U
        self.Q = self.I
        self.is_initialized = True

    def _get_topk_recommendation(self, rows, topk, pool=None):
        u = self.U[rows]
        topks = super()._get_topk_recommendation(u, self.I, pool, topk, self.opt.num_workers)
        return zip(rows, topks)

    def _get_most_similar_item(self, col, topk, pool):
        return super()._get_most_similar_item(col, topk, self.I, self.opt._nrz_I, pool)

    def get_scores(self, row_col_pairs):
        rets = {(r, c): self.U[r].dot(self.I[c]) for r, c in row_col_pairs}
        return rets

    def _get_buffer(self):
        buf = BufferedDataMatrix()
        buf.initialize(self.data, with_sppmi=True)
        return buf

    def _iterate(self, buf, group='user'):
        assert group in ["user", "item", "context"], f"group {group} is not properly provided"
        header = self.data.get_scale_info(with_sppmi=True)
        err, update_t, feed_t, updated = 0, 0, 0, 0
        if group == "user":
            self.obj.precompute("item".encode("utf8"))
            total = header["num_nnz"]
            _groups = ["rowwise"]
        elif group == "item":
            self.obj.precompute("user".encode("utf8"))
            total = header["num_nnz"] + header["sppmi_nnz"]
            _groups = ["colwise", "sppmi"]
        elif group == "context":
            total = header["sppmi_nnz"]
            _groups = ["sppmi"]

        with log.pbar(log.DEBUG, desc='%s' % group,
                      total=total, mininterval=30) as pbar:
            st = time.time()
            for start_x, next_x in buf.fetch_batch_range(_groups):
                feed_t += time.time() - st
                _err, _updated, _update_t, _feed_t = \
                    self.partial_update(buf, group, start_x, next_x)
                update_t += _update_t
                updated += _updated
                err += _err
                pbar.update(_updated)
                st = time.time()
            pbar.refresh()
        self.logger.debug(f'updated {group} processed({updated}) elapsed(data feed: {feed_t:.3f} update: {update_t:.3f}")')
        return err

    def partial_update(self, buf, group, start_x, next_x):
        st = time.time()
        if group == "user":
            indptr, keys, vals = buf.get_specific_chunk("rowwise", start_x, next_x)
            feed_t, st = time.time() - st, time.time()
            err = self.obj.partial_update_user(start_x, next_x, indptr, keys, vals)
            return err, len(keys), time.time() - st, feed_t
        elif group == "item":
            indptr_u, keys_u, vals_u = buf.get_specific_chunk("colwise", start_x, next_x)
            indptr_c, keys_c, vals_c = buf.get_specific_chunk("sppmi", start_x, next_x)
            feed_t, st = time.time() - st, time.time()
            err =  self.obj.partial_update_item(start_x, next_x, indptr_u, keys_u, vals_u,
                                                 indptr_c, keys_c, vals_c)
            return err, len(keys_u) + len(keys_c), time.time() - st, feed_t
        elif group == "context":
            indptr, keys, vals = buf.get_specific_chunk("sppmi", start_x, next_x)
            feed_t, st = time.time() - st, time.time()
            err = self.obj.partial_update_context(start_x, next_x, indptr, keys, vals)
            return err, len(keys), time.time() - st, feed_t

    def compute_scale(self):
        # scaling loss for convenience in monitoring
        ret = self.data.get_scale_info(with_sppmi=True)
        num_users, num_items, sppmi_nnz, vsum = \
            [ret[k] for k in ["num_users", "num_items", "sppmi_nnz", "vsum"]]
        alpha, l = self.opt.alpha, self.opt.l
        return l * (alpha * vsum + num_users * num_items) + sppmi_nnz

    def train(self):
        assert self.is_initialized, "embedding matrix is not initialized"
        buf = self._get_buffer()
        best_loss, rmse, self.validation_result = 987654321.0, None, {}
        self.prepare_evaluation()
        self.initialize_tensorboard(self.opt.num_iters)
        scale = self.compute_scale()
        for i in range(self.opt.num_iters):
            start_t = time.time()
            loss = self._iterate(buf, group='user')
            loss += self._iterate(buf, group='item')
            loss += self._iterate(buf, group='context')
            loss /= scale
            train_t = time.time() - start_t
            metrics = {'train_loss': loss}
            if self.opt.validation and self.opt.evaluation_on_learning and self.periodical(self.opt.evaluation_period, i):
                start_t = time.time()
                self.validation_result = self.get_validation_results()
                vali_t = time.time() - start_t
                val_str = ' '.join([f'{k}:{v:0.5f}' for k, v in self.validation_result.items()])
                self.logger.info(f'Validation: {val_str} Elased {vali_t:0.3f}')
                metrics.update({'vali_%s' % k: v
                                for k, v in self.validation_result.items()})
            self.logger.info('Iteration %d: Loss %.3f Elapsed %.3f secs' % (i + 1, loss, train_t))
            self.update_tensorboard_data(metrics)
            best_loss = self.save_best_only(loss, best_loss, i)
            if self.early_stopping(loss):
                break
        ret = {'train_loss': loss}
        ret.update({'vali_%s' % k: v
                    for k, v in self.validation_result.items()})
        self.finalize_tensorboard()
        return ret

    def _optimize(self, params):
        self._optimize_params = params
        for name, value in params.items():
            assert name in self.opt, 'Unexepcted parameter: {}'.format(name)
            setattr(self.opt, name, value)
        with open(self._temporary_opt_file, 'w') as fout:
            json.dump(self.opt, fout, indent=2)
        assert self.obj.init(bytes(self._temporary_opt_file, 'utf-8')),\
            'cannot parse option file: %s' % self._temporary_opt_file
        self.logger.info(params)
        self.initialize()
        loss = self.train()
        loss['eval_time'] = time.time()
        loss['loss'] = loss.get(self.opt.optimize.loss)
        # TODO: deal with failture of training
        loss['status'] = HOPT_STATUS_OK
        self._optimize_loss = loss
        return loss

    def _get_feature(self, index, group='item'):
        if group == 'item':
            return self.I[index]
        elif group == 'user':
            return self.U[index]
        elif group == 'context':
            return self.C[index]
        return None

    def _get_data(self):
        data = super()._get_data()
        data.extend([('opt', self.opt),
                     ('I', self.I),
                     ('U', self.U),
                     ('C', self.C)])
        return data

    def get_evaluation_metrics(self):
        return ['train_loss', 'vali_rmse', 'vali_ndcg', 'vali_map', 'vali_accuracy', 'vali_error']



# -*- coding: utf-8 -*-
import time
import json

import numpy as np
from hyperopt import STATUS_OK as HOPT_STATUS_OK

import buffalo.data
from buffalo.misc import aux, log
from buffalo.data.base import Data
from buffalo.algo._als import CyALS
from buffalo.evaluate import Evaluable
from buffalo.algo.options import ALSOption
from buffalo.algo.optimize import Optimizable
from buffalo.algo.tensorflow._als import TFALS
from buffalo.data.buffered_data import BufferedDataMatrix
from buffalo.algo.base import Algo, Serializable, TensorboardExtention

inited_CUALS = True
try:
    from buffalo.algo.cuda._als import CyALS as CuALS
except Exception as e:
    inited_CUALS = False


class ALS(Algo, ALSOption, Evaluable, Serializable, Optimizable, TensorboardExtention):
    """Python implementation for C-ALS.

    Implementation of Collaborative Filtering for Implicit Feedback datasets.

    Reference: http://yifanhu.net/PUB/cf.pdf"""
    def __init__(self, opt_path=None, *args, **kwargs):
        Algo.__init__(self, *args, **kwargs)
        ALSOption.__init__(self, *args, **kwargs)
        Evaluable.__init__(self, *args, **kwargs)
        Serializable.__init__(self, *args, **kwargs)
        Optimizable.__init__(self, *args, **kwargs)
        if opt_path is None:
            opt_path = ALSOption().get_default_option()

        self.logger = log.get_logger('ALS')
        self.opt, self.opt_path = self.get_option(opt_path)
        self.obj = CuALS() if self.opt.accelerator else CyALS()
        if self.opt.accelerator and not inited_CUALS:
            self.logger.error("ImportError CuALS, no cuda library exists.")
            raise RuntimeError()
        assert self.obj.init(bytes(self.opt_path, 'utf-8')),\
            'cannot parse option file: %s' % opt_path
        self.data = None
        data = kwargs.get('data')
        data_opt = self.opt.get('data_opt')
        data_opt = kwargs.get('data_opt', data_opt)
        if data_opt:
            self.data = buffalo.data.load(data_opt)
            self.data.create()
        elif isinstance(data, Data):
            self.data = data
        self.logger.info('ALS(%s)' % json.dumps(self.opt, indent=2))
        if self.data:
            self.logger.info(self.data.show_info())
            assert self.data.data_type in ['matrix']

    @staticmethod
    def new(path, data_fields=[]):
        return ALS.instantiate(ALSOption, path, data_fields)

    def set_data(self, data):
        assert isinstance(data, aux.data.Data), 'Wrong instance: {}'.format(type(data))
        self.data = data

    def normalize(self, group='item'):
        if group == 'item' and not self.opt._nrz_Q:
            self.Q = self._normalize(self.Q)
            self.opt._nrz_Q = True
        elif group == 'user' and not self.opt._nrz_P:
            self.P = self._normalize(self.P)
            self.opt._nrz_P = True

    def initialize(self):
        super().initialize()
        self.init_factors()

    def init_factors(self):
        assert self.data, 'Data is not setted'
        self.vdim = self.obj.get_vdim() if self.opt.accelerator else self.opt.d
        header = self.data.get_header()
        for name, rows in [('P', header['num_users']), ('Q', header['num_items'])]:
            setattr(self, name, None)
            setattr(self, name, np.abs(np.random.normal(scale=1.0/(self.opt.d ** 2),
                                       size=(rows, self.vdim)).astype("float32")))
        self.P[:, self.opt.d:] = 0.0
        self.Q[:, self.opt.d:] = 0.0
        self.obj.initialize_model(self.P, self.Q)

    def _get_topk_recommendation(self, rows, topk, pool=None):
        p = self.P[rows]
        topks = super()._get_topk_recommendation(p, self.Q, pool, topk, self.opt.num_workers)
        return zip(rows, topks)

    def _get_most_similar_item(self, col, topk, pool):
        return super()._get_most_similar_item(col, topk, self.Q, self.opt._nrz_Q, pool)

    def get_scores(self, row_col_pairs):
        rets = {(r, c): self.P[r].dot(self.Q[c]) for r, c in row_col_pairs}
        return rets

    def _get_buffer(self):
        buf = BufferedDataMatrix()
        buf.initialize(self.data)
        return buf

    def _iterate(self, buf, group='rowwise'):
        header = self.data.get_header()
        # end = header['num_users'] if group == 'rowwise' else header['num_items']
        int_group = 0 if group == 'rowwise' else 1
        st = time.time()
        self.obj.precompute(int_group)
        el, st = time.time() - st, time.time()
        loss_nume, loss_deno = 0.0, 0.0
        update_t, feed_t, updated = el, 0, 0
        buf.set_group(group)
        with log.pbar(log.DEBUG, desc='%s' % group,
                      total=header['num_nnz'], mininterval=30) as pbar:
            for sz in buf.fetch_batch():
                updated += sz
                start_x, next_x, indptr, keys, vals = buf.get()
                _feed_t, st = time.time() - st, time.time()

                _loss_nume, _loss_deno = self.obj.partial_update(start_x, next_x, indptr, keys, vals, int_group)
                loss_nume += _loss_nume
                loss_deno += _loss_deno

                _update_t, st = time.time() - st, time.time()
                pbar.update(sz)
                feed_t += _feed_t
                update_t += _update_t
        self.logger.debug(f'{group} updated: processed({updated}) elapsed(data feed: {feed_t:0.3f}s update: {update_t:0.03}s)')
        return loss_nume, loss_deno

    def train(self):
        if self.opt.accelerator:
            for attr in ["P", "Q"]:
                F = getattr(self, attr)
                if F.shape[1] < self.vdim:
                    _F = np.empty(shape=(F.shape[0], self.vdim), dtype=np.float32)
                    _F[:, :self.P.shape[1]] = F
                    _F[:, self.opt.d: ] = 0.0
                    setattr(self, attr, _F)
            self.obj.initialize_model(self.P, self.Q)

        buf = self._get_buffer()
        if self.opt.accelerator:
            lindptr, rindptr, batch_size = buf.get_indptrs()
            self.obj.set_placeholder(lindptr, rindptr, batch_size)

        best_loss, rmse, self.validation_result = 987654321.0, None, {}
        self.prepare_evaluation()
        self.initialize_tensorboard(self.opt.num_iters)
        full_st = time.time()
        for i in range(self.opt.num_iters):
            start_t = time.time()

            _loss_nume1, _loss_deno1 = self._iterate(buf, group='rowwise')
            _loss_nume2, _loss_deno2 = self._iterate(buf, group='colwise')
            loss_nume = _loss_nume1 + _loss_nume2
            loss_deno = _loss_deno1 + _loss_deno2

            train_t = time.time() - start_t
            rmse = (loss_nume / (loss_deno + self.opt.eps)) ** 0.5
            metrics = {'train_loss': rmse}
            if self.opt.validation and self.opt.evaluation_on_learning and self.periodical(self.opt.evaluation_period, i):
                start_t = time.time()
                self.validation_result = self.get_validation_results()
                vali_t = time.time() - start_t
                val_str = ' '.join([f'{k}:{v:0.5f}' for k, v in self.validation_result.items()])
                self.logger.info(f'Validation: {val_str} Elapsed {vali_t:0.3f} secs')
                metrics.update({'val_%s' % k: v
                                for k, v in self.validation_result.items()})
            self.logger.info('Iteration %d: RMSE %.3f Elapsed %.3f secs' % (i + 1, rmse, train_t))
            self.update_tensorboard_data(metrics)
            best_loss = self.save_best_only(rmse, best_loss, i)
            if self.early_stopping(rmse):
                break
        full_el = time.time() - full_st
        self.logger.info(f"elapsed for full epochs: {full_el:.2f} sec")
        if self.opt.accelerator and self.opt.d < self.vdim:
            self.P = self.P[:, :self.opt.d]
            self.Q = self.Q[:, :self.opt.d]
        ret = {'train_loss': rmse}
        ret.update({'val_%s' % k: v
                    for k, v in self.validation_result.items()})
        self.finalize_tensorboard()
        return ret

    def _optimize(self, params):
        self._optimize_params = params
        for name, value in params.items():
            assert name in self.opt, 'Unexepcted parameter: {}'.format(name)
            setattr(self.opt, name, value)
        with open(self._temporary_opt_file, 'w') as fout:
            json.dump(self.opt, fout, indent=2)
        assert self.obj.init(bytes(self._temporary_opt_file, 'utf-8')),\
            'cannot parse option file: %s' % self._temporary_opt_file
        self.logger.info(params)
        self.initialize()
        loss = self.train()
        loss['eval_time'] = time.time()
        loss['loss'] = loss.get(self.opt.optimize.loss)
        if any([metric in self.opt.optimize.loss for metric in ['ndcg', 'map', 'accracy']]):
            loss['loss'] *= -1
        # TODO: deal with failture of training
        loss['status'] = HOPT_STATUS_OK
        self._optimize_loss = loss
        return loss

    def _get_feature(self, index, group='item'):
        if group == 'item':
            return self.Q[index]
        elif group == 'user':
            return self.P[index]
        return None

    def _get_data(self):
        data = super()._get_data()
        data.extend([('opt', self.opt),
                     ('Q', self.Q),
                     ('P', self.P)])
        return data

    def get_evaluation_metrics(self):
        return ['train_loss', 'val_rmse', 'val_ndcg', 'val_map', 'val_accuracy', 'val_error']



# -*- coding: utf-8 -*-
import os
import time
import shutil

import fire

from buffalo.misc import log, aux
from buffalo.algo.als import ALS as _ALS


class ALS:
    def __init__(self):
        self.logger = log.get_logger('ALS')

    def run(self, opt_path):
        opt = aux.Option(opt_path)
        als = _ALS(opt_path)
        als.init_factors()
        loss = als.train()
        self.logger.info(f'ALS finished with loss({loss}).')
        if opt.save_factors:
            self.logger.info(f'Saving model to {opt.model_path}.')
            als.dump(opt.model_path)

    def optimize(self, opt_path):
        als = _ALS(opt_path)
        als.init_factors()
        als.optimize()
        optimize_loss = als.get_optimization_data()['best']
        self.logger.info(f'ALS optimization is done with best loss({optimize_loss})')


def execute(algo_name, method, opt_path):
    algo_cls = globals()[algo_name]
    algo = algo_cls()
    getattr(algo, method)(opt_path)


def run(algo_name, opt_path):
    execute(algo_name, 'run', opt_path)


def optimize(algo_name, opt_path):
    execute(algo_name, 'optimize', opt_path)


def _cli_buffalo():
    fire.Fire({'run': run,
               'optimize': optimize})



# cython: experimental_cpp_class_def=True
# distutils: language=c++
# -*- coding: utf-8 -*-
import io
import os
import sys
import logging
import logging.handlers
from contextlib import contextmanager

import tqdm

NOTSET = 0
WARN = 1
INFO = 2
DEBUG = 3
TRACE = 4
__logger_holder = []


cdef extern from "buffalo/misc/log.hpp":
    cdef cppclass _BuffaloLogger "BuffaloLogger":
        void set_log_level(int)
        int get_log_level()


cdef class PyBuffaloLog:
    """CALS object holder"""
    cdef _BuffaloLogger* obj

    def __cinit__(self):
        self.obj = new _BuffaloLogger()

    def __dealloc__(self):
        del self.obj

    def set_log_level(self, lvl):
        self.obj.set_log_level(lvl)

    def get_log_level(self):
        return self.obj.get_log_level()


def set_log_level(lvl):
    for logger in __logger_holder:
        if lvl == NOTSET:
            logger.setLevel(logging.NOTSET)
        elif lvl == WARN:
            logger.setLevel(logging.WARN)
        elif lvl == INFO:
            logger.setLevel(logging.INFO)
        elif lvl == DEBUG:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.DEBUG)
    PyBuffaloLog().set_log_level(lvl)


def get_log_level():
    return PyBuffaloLog().get_log_level()


class TqdmLogger(io.StringIO):
    logger = None
    buf = ''

    def __init__(self, logger):
        super(TqdmLogger, self).__init__()
        self.logger = logger

    def write(self,buf):
        self.buf = buf.strip('\r\n\t ')

    def flush(self):
        if self.buf.strip():
            self.logger(self.buf)
            self.buf = ''


def get_logger(name=__file__, no_fileno=False):
    global __logger_holder
    logger = logging.getLogger(name)
    if logger.handlers:
        return logger

    lvl = get_log_level()
    if lvl == NOTSET:
        logger.setLevel(logging.NOTSET)
    elif lvl == WARN:
        logger.setLevel(logging.WARN)
    elif lvl == INFO:
        logger.setLevel(logging.INFO)
    elif lvl == DEBUG:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.DEBUG)

    sh = logging.StreamHandler()
    sh.setLevel(logging.DEBUG)
    if no_fileno:
        formatter = logging.Formatter('[%(levelname)-8s] %(asctime)s %(message)s', '%Y-%m-%d %H:%M:%S')
    else:
        formatter = logging.Formatter('[%(levelname)-8s] %(asctime)s [%(filename)s:%(lineno)d] %(message)s', '%Y-%m-%d %H:%M:%S')
    sh.setFormatter(formatter)

    logger.addHandler(sh)
    __logger_holder.append(logger)
    return logger


@contextmanager
def pbar(log_level=INFO, **kwargs):
    if 'total' not in kwargs:
        fmt = '{desc}: {n_fmt}/{total_fmt} {elapsed}<{remaining}'
    else:
        fmt = '{l_bar} {elapsed}<{remaining}'
    if 'bar_format' not in kwargs:
        kwargs['bar_format'] = fmt
    if 'desc' not in kwargs:
        kwargs['desc'] = 'progress'
    if 'mininterval' not in kwargs:
        kwargs['mininterval'] = 1
    kwargs['leave'] = False
    logger = get_logger('pbar', no_fileno=True)
    logger_func = logger.info
    if log_level == INFO:
        logger_func = logger.info
    elif log_level == WARN:
        logger_func = logger.warn
    elif log_level == DEBUG:
        logger_func = logger.debug
    yield tqdm.tqdm(file=TqdmLogger(logger_func), **kwargs)


def iter_pbar(log_level=INFO, **kwargs):
    if 'total' not in kwargs:
        fmt = '{desc}: {n_fmt}/{total_fmt} {elapsed}<{remaining}'
    else:
        fmt = '{l_bar} {elapsed}<{remaining}'
    if 'bar_format' not in kwargs:
        kwargs['bar_format'] = fmt
    if 'desc' not in kwargs:
        kwargs['desc'] = 'progress'
    if 'mininterval' not in kwargs:
        kwargs['mininterval'] = 1
    kwargs['leave'] = False
    logger = get_logger('pbar', no_fileno=True)
    logger_func = logger.info
    if log_level == INFO:
        logger_func = logger.info
    elif log_level == WARN:
        logger_func = logger.warn
    elif log_level == DEBUG:
        logger_func = logger.debug
    return tqdm.tqdm(file=TqdmLogger(logger_func), **kwargs)


class supress_log_level(object):
    def __init__(self, log_level):
        self.desire_log_level = log_level
        self.original_log_level = get_log_level()

    def __enter__(self):
        set_log_level(self.desire_log_level)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_log_level(self.original_log_level)



# -*- coding: utf-8 -*-
import abc
import bisect
import numpy as np
from buffalo.misc import log


class BufferedData(object):
    def __init__(self):
        self.logger = log.get_logger('BufferedData')

    @abc.abstractmethod
    def initialize(self, limit):
        pass

    @abc.abstractmethod
    def reset(self):
        self.index = 0
        self.ptr_index = 0

    @abc.abstractmethod
    def get(self):
        pass


class BufferedDataMatrix(BufferedData):
    """Buffered Data for MatrixMarket

    This class feed chunked data to training step.
    """
    def __init__(self):
        super().__init__()
        self.group = 'rowwise'
        self.major = {'rowwise': {}, 'colwise': {}, 'sppmi': {}}

    def free(self, buf):
        buf['indptr'] = None
        buf['keys'] = None
        buf['vals'] = None

    def get_indptrs(self):
        return (self.major["rowwise"]["indptr"],
                self.major["colwise"]["indptr"],
                self.major["rowwise"]["limit"])

    def initialize(self, data, with_sppmi=False):
        self.data = data
        # 16 bytes(indptr8, keys4, vals4)
        limit = max(int(((self.data.opt.data.batch_mb * 1024 * 1024) / 16.)), 64)
        minimum_required_batch_size = 0
        Gs = ['rowwise', 'colwise']
        if with_sppmi:
            Gs.append('sppmi')
        for G in Gs:
            lim = int(limit / 2)
            group = data.get_group(G)
            header = data.get_header()
            if self.major[G]:
                self.free(self.major[G])
            m = self.major[G]
            m['index'] = 0
            m['limit'] = lim
            m['start_x'] = 0
            m['next_x'] = 0
            m['max_x'] = header['num_users'] if G == 'rowwise' else header['num_items']
            m['indptr'] = group['indptr'][::]
            minimum_required_batch_size = max([m['indptr'][i] - m['indptr'][i - 1]
                                               for i in range(1, len(m['indptr']))])
            m['keys'] = np.zeros(shape=(lim,), dtype=np.int32, order='C')
            m['vals'] = np.zeros(shape=(lim,), dtype=np.float32, order='C')
        self.logger.info(f'Set data buffer size as {limit}(minimum required batch size is {minimum_required_batch_size}).')
        if minimum_required_batch_size > int(limit / 2):
            self.logger.warning('Given batch size(%d) is smaller than '
                                'minimum required batch size(%d) for the data. '
                                'Increasing batch_mb would be helpful for faster traininig.',
                                int(limit / 2), minimum_required_batch_size)
            for G in ['rowwise', 'colwise']:
                m = self.major[G]
                lim = minimum_required_batch_size + 1
                m['limit'] = lim
                m['keys'] = np.zeros(shape=(lim,), dtype=np.int32, order='C')
                m['vals'] = np.zeros(shape=(lim,), dtype=np.float32, order='C')

    def fetch_batch(self):
        m = self.major[self.group]
        flushed = False
        while True:
            if m['start_x'] == 0 and m['next_x'] + 1 >= m['max_x']:
                if not flushed:
                    m["sz"] = m["indptr"][-1]
                    yield m['indptr'][-1]
                return

            if m['next_x'] + 1 >= m['max_x']:
                m['start_x'], m['next_x'] = 0, 0
                return

            m['start_x'] = m['next_x']

            group = self.data.get_group(self.group)
            beg = 0 if m['start_x'] == 0 else m['indptr'][m['start_x'] - 1]
            where = bisect.bisect_left(m['indptr'], beg + m['limit'])
            if where == m['start_x']:
                current_batch_size = m['limit']
                need_batch_size = m['indptr'][where] - beg
                raise RuntimeError('Need more memory to load the data, '
                                   'cannot load data with buffer size %d that should be at least %d. '
                                   'Increase batch_mb value to deal with this.' % (current_batch_size, need_batch_size))
            end = m['indptr'][where - 1]
            m['next_x'] = where
            size = end - beg
            m['keys'][:size] = group['key'][beg:end]
            m['vals'][:size] = group['val'][beg:end]
            if m['next_x'] + 1 >= m['max_x']:
                flushed = True
            m["sz"] = size
            yield size

    def get_specific_chunk(self, group, start_x, next_x):
        db = self.data.get_group(group)
        m = self.major[group]
        indptr = m["indptr"]
        beg = 0 if start_x == 0 else indptr[start_x - 1]
        end = indptr[next_x - 1]
        keys = db["key"][beg: end]
        vals = db["val"][beg: end]
        return indptr, keys, vals

    def fetch_batch_range(self, groups):
        assert groups and isinstance(groups, list), "groups should be list (length >= 1)"
        for G in groups:
            assert G in ["rowwise", "colwise", "sppmi"], f"G ({G}) is not proper group for getting range"
        db = self.data.get_group(groups[0])
        indptr = np.zeros(db["indptr"].shape[0], dtype=np.int64)
        max_x = len(indptr)
        for G in groups:
            _indptr = self.data.get_group(G)["indptr"][:]
            assert len(_indptr) == max_x, f"size of indptr (group: {G}) should be {max_x}"
            indptr += _indptr
        limit = max(int(((self.data.opt.data.batch_mb * 1024 * 1024) / 8.)), 64)
        start_x, next_x, flushed = 0, 0, False
        while True:
            if flushed:
                return

            start_x = next_x

            beg = 0 if start_x == 0 else indptr[start_x - 1]
            where = bisect.bisect_left(indptr, beg + limit)
            if where == start_x:
                current_batch_size = limit
                need_batch_size = indptr[where] - beg
                raise RuntimeError('Need more memory to load the data, '
                                   'cannot load data with buffer size %d that should be at least %d. '
                                   'Increase batch_mb value to deal with this.' % (current_batch_size, need_batch_size))
            next_x = where
            if next_x + 1 >= max_x:
                flushed = True
            yield start_x, next_x

    def set_group(self, group):
        assert group in ['rowwise', 'colwise', 'sppmi'], 'Unexpected group: {}'.format(group)
        self.group = group

    def reset(self):
        for m in self.major.valus():
            m['index'], m['start_x'], m['next_x'] = 0, 0, 0

    def get(self):
        m = self.major[self.group]
        return [m[k] for k in ["start_x", "next_x", "indptr", "keys", "vals"]]


class BufferedDataStream(BufferedData):
    """Buffered Data for Stream

    This class feed chunked data to training step.
    """
    def __init__(self):
        super().__init__()
        self.major = {'rowwise': {}}
        self.group = 'rowwise'

    def free(self, buf):
        buf['indptr'] = None
        buf['keys'] = None

    def initialize(self, data):
        self.data = data
        assert self.data.data_type == 'stream'
        # 12 bytes(indptr8, keys4)
        limit = max(int(((self.data.opt.data.batch_mb * 1000 * 1000) / 12.)), 64)
        minimum_required_batch_size = 0
        lim = int(limit / 2)
        G = 'rowwise'
        group = data.get_group(G)
        header = data.get_header()
        m = self.major[G]
        if self.major[G]:
            self.free(self.major[G])
        m['index'] = 0
        m['limit'] = lim
        m['start_x'] = 0
        m['next_x'] = 0
        m['max_x'] = header['num_users']
        m['indptr'] = group['indptr'][::]
        minimum_required_batch_size = max([m['indptr'][i] - m['indptr'][i - 1]
                                           for i in range(1, len(m['indptr']))])
        m['keys'] = np.zeros(shape=(lim,), dtype=np.int32)
        if minimum_required_batch_size > int(limit / 2):
            self.logger.warning('Given batch size(%d) is smaller than '
                                'minimum required batch size(%d) is for the data. '
                                'Increasing batch_mb would be helpful for faster traininig.',
                                int(limit / 2), minimum_required_batch_size)
            m = self.major[G]
            lim = minimum_required_batch_size + 1
            m['limit'] = lim
            m['keys'] = np.zeros(shape=(lim,), dtype=np.int32)

    def fetch_batch(self):
        m = self.major[self.group]
        flushed = False
        while True:
            if m['start_x'] == 0 and m['next_x'] + 1 >= m['max_x']:
                if not flushed:
                    yield m['indptr'][-1]
                return

            if m['next_x'] + 1 >= m['max_x']:
                m['start_x'], m['next_x'] = 0, 0
                return

            m['start_x'] = m['next_x']

            group = self.data.get_group(self.group)
            beg = 0 if m['start_x'] == 0 else m['indptr'][m['start_x'] - 1]
            where = bisect.bisect_left(m['indptr'], beg + m['limit'])
            if where == m['start_x']:
                current_batch_size = m['limit']
                need_batch_size = m['indptr'][where] - beg
                raise RuntimeError('Need more memory to load the data, '
                                   'cannot load data with buffer size %d that should be at least %d. '
                                   'Increase batch_mb value to deal with this.' % (current_batch_size, need_batch_size))
            end = m['indptr'][where - 1]
            m['next_x'] = where
            size = end - beg
            m['keys'][:size] = group['key'][beg:end]
            if m['next_x'] + 1 >= m['max_x']:
                flushed = True
            yield size

    def set_group(self, group):
        assert group in ['rowwise', 'sppmi'], 'Unexpected group: {}'.format(group)
        self.group = group

    def reset(self):
        for m in self.major.valus():
            m['index'], m['start_x'], m['next_x'] = 0, 0, 0

    def get(self):
        m = self.major[self.group]
        return (m['start_x'],
                m['next_x'],
                m['indptr'],
                m['keys'])



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import time
import json

import numpy as np
from hyperopt import STATUS_OK as HOPT_STATUS_OK

import buffalo.data
from buffalo.data.base import Data
from buffalo.misc import aux, log
from buffalo.algo._w2v import CyW2V
from buffalo.evaluate import Evaluable
from buffalo.algo.options import W2VOption
from buffalo.algo.optimize import Optimizable
from buffalo.data.buffered_data import BufferedDataStream
from buffalo.algo.base import Algo, Serializable, TensorboardExtention


class W2V(Algo, W2VOption, Evaluable, Serializable, Optimizable, TensorboardExtention):
    """Python implementation for C-W2V
    """
    def __init__(self, opt_path=None, *args, **kwargs):
        Algo.__init__(self, *args, **kwargs)
        W2VOption.__init__(self, *args, **kwargs)
        Evaluable.__init__(self, *args, **kwargs)
        Serializable.__init__(self, *args, **kwargs)
        Optimizable.__init__(self, *args, **kwargs)
        if opt_path is None:
            opt_path = W2VOption().get_default_option()

        self.logger = log.get_logger('W2V')
        self.opt, self.opt_path = self.get_option(opt_path)
        self.obj = CyW2V()
        assert self.obj.init(bytes(self.opt_path, 'utf-8')),\
            'cannot parse option file: %s' % opt_path
        self.data = None
        data = kwargs.get('data')
        data_opt = self.opt.get('data_opt')
        data_opt = kwargs.get('data_opt', data_opt)
        if data_opt:
            self.data = buffalo.data.load(data_opt)
            assert self.data.data_type == 'stream'
            self.data.create()
        elif isinstance(data, Data):
            self.data = data
        self.logger.info('W2V(%s)' % json.dumps(self.opt, indent=2))
        if self.data:
            self.logger.info(self.data.show_info())
            assert self.data.data_type in ['stream']
        self._vocab = aux.Option({'size': 0,
                                  'index': None,
                                  'inv_index': None,
                                  'scale': None,
                                  'dist': None,
                                  'total_word_count': 0})

    @staticmethod
    def new(path, data_fields=[]):
        return W2V.instantiate(W2VOption, path, data_fields)

    def set_data(self, data):
        assert isinstance(data, aux.data.Data), 'Wrong instance: {}'.format(type(data))
        self.data = data

    def normalize(self, group='item'):
        if group == 'item' and not self.opt._nrz_L0:
            self.L0 = self._normalize(self.L0)
            self.opt._nrz_L0 = True

    def get_index(self, key, group='item'):
        is_many = isinstance(key, list)
        indexes = super().get_index(key, group)
        if not is_many:
            indexes = [indexes]
        indexes = [i if i is None or self._vocab.index[i] < 1 else self._vocab.index[i] - 1
                   for i in indexes]
        if not is_many:
            return indexes[0]
        return np.array(indexes)

    def _get_feature(self, index, group='item'):
        if group == 'item':
            return self.L0[index]
        return None

    def initialize(self):
        super().initialize()
        assert self.data, 'Data is not setted'
        if self.opt.random_seed:
            np.random.seed(self.opt.random_seed)
        self.buf = BufferedDataStream()
        self.buf.initialize(self.data)
        self.build_vocab()
        self.init_factors(self._vocab.size)
        self.obj.initialize_model(self.L0, self._vocab.index, self._vocab.scale,
                                  self._vocab.dist, self._vocab.total_word_count)

    def build_vocab(self):
        header = self.data.get_header()
        self.logger.info('Caculating the frequency of words.')
        uni = [0 for i in range(header['num_items'])]
        total_word_count = 0
        for sz in self.buf.fetch_batch():
            *_, keys = self.buf.get()
            for i in range(sz):
                uni[keys[i]] += 1
            total_word_count += sz
        self.logger.info(f'Reducing vocab with min_count({self.opt.min_count}).')
        total_vocab = 0
        # use = np.zeros(shape=header['num_items'], dtype=np.int32, order='F')
        use = [0 for i in range(header['num_items'])]
        for i in range(header['num_items']):
            if uni[i] >= self.opt.min_count:
                total_vocab += 1
                use[i] = total_vocab
        self.logger.info(f'Scaling vocab({total_vocab}) with sample({self.opt.sample}).')
        scale = np.zeros(shape=total_vocab, dtype=np.uint32, order='C')
        threshold_count = sum([uni[i] for i in range(header['num_items']) if use[i]])
        if self.opt.sample > 0.0:
            threshold_count *= self.opt.sample
        num_downsampled = 0
        for i in range(header['num_items']):
            if not use[i]:
                continue
            p = (((uni[i] / threshold_count) ** 0.5) + 1) * (threshold_count / uni[i])
            if p < 1.0:
                num_downsampled += 1
            else:
                p = 1.0
            scale[use[i] - 1] = p * 0xFFFFFFFF
        self.logger.info(f'Downsampled {num_downsampled} most-common words.')
        dist = self.get_sampling_distribution(uni, use, total_vocab)
        self._vocab.size = total_vocab
        self._vocab.scale = scale
        self._vocab.index = np.array(use, dtype=np.int32, order='C')
        self._vocab.inv_index = np.array([idx for idx, u in enumerate(use) if u > 0],
                                         dtype=np.int32)
        self._vocab.dist = dist
        self._vocab.total_word_count = total_word_count
        self.logger.info(f'Vocab({total_vocab}) TotalWords({total_word_count})')

    def init_factors(self, vocab_size):
        self.L0 = None
        self.L0 = np.abs(np.random.normal(scale=1.0/(self.opt.d ** 2),
                                         size=(vocab_size, self.opt.d)).astype("float32"))

    def get_sampling_distribution(self, uni, use, total_vocab):
        dist0 = np.zeros(shape=total_vocab, dtype=np.float64, order='C')
        for i in range(len(use)):
            if not use[i]:
                continue
            dist0[use[i] - 1] = uni[i]
        dist0 = dist0 ** 0.75
        train_word_pow = dist0.sum()
        dist0 /= train_word_pow

        dist = np.zeros(shape=total_vocab, dtype=np.int32, order='C')
        summed = 0.0
        domain = 0x7FFFFFFF
        for i in range(total_vocab):
            summed += dist0[i]
            dist[i] = summed * domain
        assert(dist[-1] == domain)
        return dist

    def _get_topk_recommendation(self, rows, topk, pool=None):
        raise NotImplemented

    def _get_most_similar_item(self, col, topk, pool):
        if not isinstance(col, np.ndarray):
            col = self._vocab.index[col] - 1
            if col < 0:
                return [], []
        topks, scores = super()._get_most_similar_item(col, topk, self.L0, self.opt._nrz_L0, pool)
        topks = self._vocab.inv_index[topks]
        return topks, scores

    def get_scores(self, row_col_pairs):
        return []

    def _iterate(self):
        header = self.data.get_header()
        end = header['num_users']
        update_t, feed_t, updated = 0, 0, 0
        self.buf.set_group('rowwise')
        with log.pbar(log.DEBUG,
                      total=header['num_nnz'], mininterval=15) as pbar:
            start_t = time.time()
            for sz in self.buf.fetch_batch():
                updated += sz
                feed_t += time.time() - start_t
                start_x, next_x, indptr, keys = self.buf.get()

                start_t = time.time()
                self.obj.add_jobs(start_x, next_x, indptr, keys)
                update_t += time.time() - start_t
                pbar.update(sz)
        self.logger.debug(f'processed({updated}) elapsed(data feed: {feed_t:0.3f}s update: {update_t:0.3f})s')

    def train(self):
        self.validation_result = {}
        self.prepare_evaluation()
        self.obj.launch_workers()
        for i in range(self.opt.num_iters):
            start_t = time.time()
            self._iterate()
            self.logger.info('Iteration %s: Elapsed %.3f secs' % (i + 1, time.time() - start_t))
        loss = self.obj.join()
        return {}

    def _optimize(self, params):
        # TODO: implement
        self._optimize_params = params
        for name, value in params.items():
            assert name in self.opt, 'Unexepcted parameter: {}'.format(name)
            setattr(self.opt, name, value)
        with open(self._temporary_opt_file, 'w') as fout:
            json.dump(self.opt, fout, indent=2)
        assert self.obj.init(bytes(self._temporary_opt_file, 'utf-8')),\
            'cannot parse option file: %s' % self._temporary_opt_file
        self.logger.info(params)
        self.init_factors()
        loss = self.train()
        loss['eval_time'] = time.time()
        loss['loss'] = loss.get(self.opt.optimize.loss)
        # TODO: deal with failture of training
        loss['status'] = HOPT_STATUS_OK
        self._optimize_loss = loss
        return loss

    def _get_data(self):
        data = super()._get_data()
        data.extend([('opt', self.opt),
                     ('L0', self.L0),
                     ('_vocab', self._vocab)])
        return data

    def get_evaluation_metrics(self):
        return []



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
#-*- coding: utf-8 -*-
import cython
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool as bool_t
from libc.stdint cimport int32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "_core.hpp":
    cdef void _quickselect "parallel::quickselect"(float*, int, int, int32_t*, int, bool_t, int) nogil except +

    cdef void _dot_topn "parallel::dot_topn"(int32_t*, int,
                                             float*, int, int,
                                             float*, int, int,
                                             float*, int,
                                             int32_t*, float*,
                                             int32_t*, int,
                                             int, int) nogil except +

    cdef void _ann_search "parallel::ann_search"(string,
                                                 int32_t,
                                                 bool_t,
                                                 int32_t*, int,
                                                 float*, int, int,
                                                 float*, int, int,
                                                 float*, int,
                                                 int32_t*, float*,
                                                 int32_t*, int,
                                                 int, int) nogil except +


@cython.boundscheck(False)
@cython.wraparound(False)
def quickselect(np.ndarray[np.float32_t, ndim=2] scores,
                np.ndarray[np.int32_t, ndim=2] result,
                sorted, num_threads):
    _quickselect(&scores[0, 0], scores.shape[0], scores.shape[1],
                 &result[0, 0], result.shape[1], sorted, num_threads)


@cython.boundscheck(False)
@cython.wraparound(False)
def dot_topn(np.ndarray[np.int32_t, ndim=1] indexes,
             np.ndarray[np.float32_t, ndim=2] P,
             np.ndarray[np.float32_t, ndim=2] Q,
             np.ndarray[np.float32_t, ndim=2] Qb,
             np.ndarray[np.int32_t, ndim=2]  out_keys,
             np.ndarray[np.float32_t, ndim=2]  out_scores,
             np.ndarray[np.int32_t, ndim=1] pool,
             k, num_threads):
    qb_shape = Qb.shape[0] if Qb.shape[1] != 0 else 0
    _dot_topn(&indexes[0], indexes.shape[0],
              &P[0, 0], P.shape[0], P.shape[1],
              &Q[0, 0], Q.shape[0], Q.shape[1],
              &Qb[0, 0], qb_shape,
              &out_keys[0, 0], &out_scores[0, 0],
              &pool[0], pool.shape[0],
              k, num_threads)


@cython.boundscheck(False)
@cython.wraparound(False)
def ann_search(index_path,
               ef_search,
               use_mmap,
               np.ndarray[np.int32_t, ndim=1] indexes,
               np.ndarray[np.float32_t, ndim=2] P,
               np.ndarray[np.float32_t, ndim=2] Q,
               np.ndarray[np.float32_t, ndim=2] Qb,
               np.ndarray[np.int32_t, ndim=2]  out_keys,
               np.ndarray[np.float32_t, ndim=2]  out_scores,
               np.ndarray[np.int32_t, ndim=1] pool,
               k, num_threads):
    qb_shape = Qb.shape[0] if Qb.shape[1] != 0 else 0
    _ann_search(index_path,
                ef_search,
                use_mmap,
                &indexes[0], indexes.shape[0],
                &P[0, 0], P.shape[0], P.shape[1],
                &Q[0, 0], Q.shape[0], Q.shape[1],
                &Qb[0, 0], qb_shape,
                &out_keys[0, 0], &out_scores[0, 0],
                &pool[0], pool.shape[0],
                k, num_threads)



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import time
import json

import numpy as np
from numpy.linalg import norm
from hyperopt import STATUS_OK as HOPT_STATUS_OK

import buffalo.data
from buffalo.data.base import Data
from buffalo.misc import aux, log
from buffalo.algo._bpr import CyBPRMF
from buffalo.evaluate import Evaluable
from buffalo.algo.options import BPRMFOption
from buffalo.algo.optimize import Optimizable
from buffalo.data.buffered_data import BufferedDataMatrix
from buffalo.algo.base import Algo, Serializable, TensorboardExtention


class BPRMF(Algo, BPRMFOption, Evaluable, Serializable, Optimizable, TensorboardExtention):
    """Python implementation for C-BPRMF.
    """
    def __init__(self, opt_path=None, *args, **kwargs):
        Algo.__init__(self, *args, **kwargs)
        BPRMFOption.__init__(self, *args, **kwargs)
        Evaluable.__init__(self, *args, **kwargs)
        Serializable.__init__(self, *args, **kwargs)
        Optimizable.__init__(self, *args, **kwargs)
        if opt_path is None:
            opt_path = BPRMFOption().get_default_option()

        self.logger = log.get_logger('BPRMF')
        self.opt, self.opt_path = self.get_option(opt_path)
        self.obj = CyBPRMF()
        assert self.obj.init(bytes(self.opt_path, 'utf-8')),\
            'cannot parse option file: %s' % opt_path
        self.data = None
        data = kwargs.get('data')
        data_opt = self.opt.get('data_opt')
        data_opt = kwargs.get('data_opt', data_opt)
        if data_opt:
            self.data = buffalo.data.load(data_opt)
            self.data.create()
        elif isinstance(data, Data):
            self.data = data
        self.logger.info('BPRMF(%s)' % json.dumps(self.opt, indent=2))
        if self.data:
            self.logger.info(self.data.show_info())
            assert self.data.data_type in ['matrix']

    @staticmethod
    def new(path, data_fields=[]):
        return BPRMF.instantiate(BPRMFOption, path, data_fields)

    def set_data(self, data):
        assert isinstance(data, aux.data.Data), 'Wrong instance: {}'.format(type(data))
        self.data = data

    def normalize(self, group='item'):
        if group == 'item' and not self.opt._nrz_Q:
            self.Q = self._normalize(self.Q)
            self.opt._nrz_Q = True
        elif group == 'user' and not self.opt._nrz_P:
            self.P = self._normalize(self.P)
            self.opt._nrz_P = True

    def initialize(self):
        super().initialize()
        assert self.data, 'Data is not setted'
        if self.opt.random_seed:
            np.random.seed(self.opt.random_seed)
        self.buf = BufferedDataMatrix()
        self.buf.initialize(self.data)
        self.init_factors()
        self.prepare_sampling()

    def init_factors(self):
        header = self.data.get_header()
        for attr_name in ['P', 'Q', 'Qb']:
            setattr(self, attr_name, None)
        self.P = np.abs(np.random.normal(scale=1.0/(self.opt.d ** 2),
                                         size=(header['num_users'], self.opt.d)).astype("float32"), order='C')
        self.Q = np.abs(np.random.normal(scale=1.0/(self.opt.d ** 2),
                                         size=(header['num_items'], self.opt.d)).astype("float32"), order='C')
        self.Qb = np.abs(np.random.normal(scale=1.0/(self.opt.d ** 2),
                                          size=(header['num_items'], 1)).astype("float32"), order='C')
        self.obj.initialize_model(self.P, self.Q, self.Qb, header['num_nnz'])

    def prepare_sampling(self):
        self.logger.info('Preparing sampling ...')
        header = self.data.get_header()
        self.sampling_table_ = np.zeros(header['num_items'], dtype=np.int64)
        if self.opt.sampling_power > 0.0:
            for sz in self.buf.fetch_batch():
                *_, keys, vals = self.buf.get()
                for i in range(sz):
                    self.sampling_table_[keys[i]] += 1
            self.sampling_table_ **= int(self.opt.sampling_power)
            for i in range(1, header['num_items']):
                self.sampling_table_[i] += self.sampling_table_[i - 1]
        self.obj.set_cumulative_table(self.sampling_table_, header['num_items'])

    def _get_topk_recommendation(self, rows, topk, pool=None):
        p = self.P[rows]
        topks = super()._get_topk_recommendation(p, self.Q, pool, topk, self.opt.num_workers)
        return zip(rows, topks)

    def _get_most_similar_item(self, col, topk, pool):
        return super()._get_most_similar_item(col, topk, self.Q, self.opt._nrz_Q, pool)

    def get_scores(self, row_col_pairs):
        rets = {(r, c): self.P[r].dot(self.Q[c]) + self.Qb[c][0] for r, c in row_col_pairs}
        return rets

    def sampling_loss_samples(self):
        users, positives, negatives = [], [], []
        if self.opt.compute_loss_on_training:
            self.logger.info('Sampling loss samples...')
            header = self.data.get_header()
            num_loss_samples = int(header['num_users'] ** 0.5)
            _users = np.random.choice(range(self.P.shape[0]), size=num_loss_samples, replace=False)
            users = []
            positives, negatives = [], []
            for u in _users:
                keys, *_ = self.data.get(u)
                if len(keys) == 0:
                    continue
                seen = set(keys)
                negs = np.random.choice(range(self.Q.shape[0]),
                                        size=len(seen) + 1,
                                        replace=False)
                negs = [n for n in negs if n not in seen]
                users.append(u)
                positives.append(keys[0])
                negatives.append(negs[0])
            self.logger.info('Generated %s loss samples.' % len(users))
        self._sub_samples = [
            np.array(users, dtype=np.int32, order='F'),
            np.array(positives, dtype=np.int32, order='F'),
            np.array(negatives, dtype=np.int32, order='F')
        ]

    def _get_feature(self, index, group='item'):
        if group == 'item':
            return self.Q[index]
        elif group == 'user':
            return self.P[index]
        return None

    def _iterate(self):
        header = self.data.get_header()
        end = header['num_users']
        update_t, feed_t, updated = 0, 0, 0
        self.buf.set_group('rowwise')
        with log.pbar(log.DEBUG,
                      total=header['num_nnz'], mininterval=30) as pbar:
            start_t = time.time()
            for sz in self.buf.fetch_batch():
                updated += sz
                feed_t += time.time() - start_t
                start_x, next_x, indptr, keys, vals = self.buf.get()

                start_t = time.time()
                self.obj.add_jobs(start_x, next_x, indptr, keys)
                update_t += time.time() - start_t
                pbar.update(sz)
            pbar.refresh()
        self.obj.update_parameters()
        self.logger.debug(f'updated processed({updated}) elapsed(data feed: {feed_t:0.3f} update: {update_t:0.3f}')

    def compute_loss(self):
        return self.obj.compute_loss(self._sub_samples[0],
                                     self._sub_samples[1],
                                     self._sub_samples[2])

    def train(self):
        rmse, self.validation_result = None, {}
        self.prepare_evaluation()
        self.initialize_tensorboard(self.opt.num_iters)
        self.sampling_loss_samples()
        self.obj.launch_workers()
        best_loss = 987654321.0
        for i in range(self.opt.num_iters):
            start_t = time.time()
            self._iterate()
            self.obj.wait_until_done()
            loss = self.compute_loss() if self.opt.compute_loss_on_training else 0.0
            train_t = time.time() - start_t

            metrics = {'train_loss': loss}
            if self.opt.validation and self.opt.evaluation_on_learning and self.periodical(self.opt.evaluation_period, i):
                start_t = time.time()
                self.validation_result = self.get_validation_results()
                vali_t = time.time() - start_t
                val_str = ' '.join([f'{k}:{v:0.5f}' for k, v in self.validation_result.items()])
                self.logger.info(f'Validation: {val_str} Elased {vali_t:0.3f}')
                metrics.update({'val_%s' % k: v
                                for k, v in self.validation_result.items()})
            self.logger.info('Iteration %s: PR-Loss %.3f Elapsed %.3f secs' % (i + 1, loss, time.time() - start_t))
            self.update_tensorboard_data(metrics)
            best_loss = self.save_best_only(loss, best_loss, i)
            if self.early_stopping(loss):
                break
        loss = self.obj.join()

        ret = {'train_loss': loss}
        ret.update({'val_%s' % k: v
                    for k, v in self.validation_result.items()})
        self.finalize_tensorboard()
        return ret

    def _optimize(self, params):
        # TODO: implement
        self._optimize_params = params
        for name, value in params.items():
            assert name in self.opt, 'Unexepcted parameter: {}'.format(name)
            setattr(self.opt, name, value)
        with open(self._temporary_opt_file, 'w') as fout:
            json.dump(self.opt, fout, indent=2)
        assert self.obj.init(bytes(self._temporary_opt_file, 'utf-8')),\
            'cannot parse option file: %s' % self._temporary_opt_file
        self.logger.info(params)
        self.init_factors()
        loss = self.train()
        loss['eval_time'] = time.time()
        loss['loss'] = loss.get(self.opt.optimize.loss)
        # TODO: deal with failture of training
        loss['status'] = HOPT_STATUS_OK
        self._optimize_loss = loss
        return loss

    def _get_data(self):
        data = super()._get_data()
        data.extend([('opt', self.opt),
                     ('Q', self.Q),
                     ('Qb', self.Qb),
                     ('P', self.P)])
        return data

    def get_evaluation_metrics(self):
        return ['val_rmse', 'val_ndcg', 'val_map', 'val_accuracy', 'val_error', 'train_loss']



# -*- coding: utf-8 -*-
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import abc
import json
import pickle
import struct
import logging
import datetime

import numpy as np
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
from tensorflow.keras.utils import Progbar
# what the...
import absl.logging
logging.root.removeHandler(absl.logging._absl_handler)
absl.logging._warn_preinit_stderr = False

from buffalo.misc import aux


class Algo(abc.ABC):
    def __init__(self, *args, **kwargs):
        self._idmanager = aux.Option({'userid': [], 'userid_map': {},
                                      'itemid': [], 'itemid_map': {},
                                      'userid_mapped': False, 'itemid_mapped': False})

    def get_option(self, opt_path) -> (aux.Option, str):
        if isinstance(opt_path, (dict, aux.Option)):
            opt_path = self.create_temporary_option_from_dict(opt_path)
        opt = aux.Option(opt_path)
        opt_path = opt_path
        self.is_valid_option(opt)
        return (aux.Option(opt), opt_path)

    def _normalize(self, feat):
        feat = feat / np.sqrt((feat ** 2).sum(-1) + 1e-8)[..., np.newaxis]
        return feat

    def initialize(self):
        self.__early_stopping = {'round': 0,
                                 'min_loss': 987654321}

    @abc.abstractmethod
    def normalize(self, group='item'):
        raise NotImplemented

    def _get_topk_recommendation(self, p, Q, pool, topk, num_workers):
        # Warning: This should be inherited.
        if pool is None:
            topks = self.get_topk(p.dot(Q.T), k=topk, num_threads=num_workers)
        else:
            topks = self.get_topk(p.dot(Q[pool].T), k=topk, num_threads=num_workers)
            topks = np.array([pool[t] for t in topks])
        return topks

    def topk_recommendation(self, keys, topk=10, pool=None):
        """Return TopK recommendation for each users(keys)

        :param keys: Query key(s)
        :type keys: list or str
        :param int topk: Number of recommendation
        :param pool: See the pool parameter of `most_similar`
        :rtype: dict or list
        """
        is_many = isinstance(keys, list)
        if not is_many:
            keys = [keys]
        if not self._idmanager.userid_mapped:
            self.build_userid_map()
        if not self._idmanager.itemid_mapped:
            self.build_itemid_map()
        if pool is not None:
            pool = self.get_index_pool(pool, group='item')
            if len(pool) == 0:
                return []
        rows = [self._idmanager.userid_map[k] for k in keys
                if k in self._idmanager.userid_map]
        topks = self._get_topk_recommendation(rows, topk, pool)
        if not topks:
            return []
        if is_many:
            return {self._idmanager.userids[k]: [self._idmanager.itemids[v] for v in vv]
                    for k, vv in topks}
        else:
            for k, vv in topks:
                return [self._idmanager.itemids[v] for v in vv]

    def most_similar(self, key, topk=10, group='item', pool=None):
        """Return top-k most similar items

        :param str key: Query key
        :param int topk: The number of results (default: 10)
        :param str group: Data group where to find (default: item)
        :param pool: The list of item keys to find for.
            If it is a numpy.ndarray instance then it treat as index of items and it would be helpful for calculation speed. (default: None)
        :type pool: list or numpy.ndarray
        :return: Top-k most similar items for given query.
        :rtype: list
        """
        if group == 'item':
            if not self._idmanager.itemid_mapped:
                self.build_itemid_map()
            return self._most_similar_item(key, topk, pool)
        return []

    def _get_most_similar_item(self, col, topk, Factor, nrz, pool):
        if isinstance(col, np.ndarray):
            q = col
        else:
            topk += 1
            q = Factor[col]
        if nrz:
            if pool is not None:
                dot = q.dot(Factor[pool].T)
            else:
                dot = q.dot(Factor.T)
            # topks = np.argsort(dot)[-topk:][::-1]
            topks = self.get_topk(dot, k=topk, num_threads=self.opt.num_workers)
        else:
            if pool is not None:
                dot = q.dot(Factor[pool].T)
                dot = dot / (np.linalg.norm(q) * np.linalg.norm(Factor[pool], axis=1))
            else:
                dot = q.dot(Factor.T)
                dot = dot / (np.linalg.norm(q) * np.linalg.norm(Factor, axis=1))
            # topks = np.argsort(dot)[-topk:][::-1]
            topks = self.get_topk(dot, k=topk, num_threads=self.opt.num_workers)
        scores = dot[topks]
        if pool is not None:
            topks = np.array([pool[t] for t in topks])
        return topks, scores

    def _most_similar_item(self, key, topk=10, pool=None):
        is_vector = False
        if isinstance(key, np.ndarray):
            f = key
            is_vector = True
        else:
            col = self._idmanager.itemid_map.get(key)
            if col is None:
                return []
            f = col
        if pool is not None:
            pool = self.get_index_pool(pool, group='item')
            if len(pool) == 0:
                return []
        topks, scores = self._get_most_similar_item(f, topk, pool)
        if is_vector:
            return [(self._idmanager.itemids[k], v)
                    for (k, v) in zip(topks, scores)]
        else:
            return [(self._idmanager.itemids[k], v)
                    for (k, v) in zip(topks, scores)
                    if k != f]

    def build_itemid_map(self):
        idmap = self.data.get_group('idmap')
        header = self.data.get_header()
        if idmap['cols'].shape[0] == 0:
            self._idmanager.itemids = list(map(str, list(range(header['num_items']))))
            self._idmanager.itemid_map = {str(i): i for i in range(header['num_items'])}
        else:
            self._idmanager.itemids = list(map(lambda x: x.decode('utf-8', 'ignore'), idmap['cols'][::]))
            self._idmanager.itemid_map = {v: idx
                                          for idx, v in enumerate(self._idmanager.itemids)}
        self._idmanager.itemid_mapped = True

    def build_userid_map(self):
        idmap = self.data.get_group('idmap')
        header = self.data.get_header()
        if idmap['rows'].shape[0] == 0:
            self._idmanager.userids = list(map(str, list(range(header['num_users']))))
            self._idmanager.userid_map = {str(i): i for i in range(header['num_users'])}
        else:
            self._idmanager.userids = list(map(lambda x: x.decode('utf-8', 'ignore'), idmap['rows'][::]))
            self._idmanager.userid_map = {v: idx
                                          for idx, v in enumerate(self._idmanager.userids)}
        self._idmanager.userid_mapped = True

    def get_feature(self, name, group='item'):
        index = self.get_index(name, group=group)
        if not index:
            return None
        return self._get_feature(index, group)

    @abc.abstractmethod
    def _get_feature(self, index, group='item'):
        raise NotImplemented

    def get_weighted_feature(self, weights, group='item', min_length=1):
        if isinstance(weights, dict):
            feat = [(self.get_feature(k), w) for k, w in weights.items()]
            feat = [f * w for f, w in feat if f is not None]
        elif isinstance(weights, list):
            feat = [self.get_feature(k) for k, w in weights]
        if len(feat) < min_length:
            return None
        feat = np.array(feat, dtype=np.float64).mean(axis=0)
        return (feat / np.linalg.norm(feat)).astype(np.float32)

    def periodical(self, period, current):
        if not period or (current + 1) % period == 0:
            return True
        return False

    def save_best_only(self, loss, best_loss, i):
        if self.opt.save_best and best_loss > loss and self.periodical(self.opt.save_period, i):
            self.save(self.opt.model_path)
            return loss
        return best_loss

    def early_stopping(self, loss):
        if self.opt.early_stopping_rounds < 1:
            return False
        if self.__early_stopping['min_loss'] < loss:
            self.__early_stopping['round'] += 1
        else:
            self.__early_stopping['round'] = 0
        self.__early_stopping['min_loss'] = loss
        if self.__early_stopping['round'] >= self.opt.early_stopping_rounds:
            self.logger.info('Reached at early_stopping rounds, stopping train.')
            return True
        return False

    def get_index(self, keys, group='item'):
        """Get index list of given item keys.
        If there is no index for such key, return None.

        :param keys: Query key(s)
        :type keys: str or list
        :param str group: Data group where to find (default: item)
        :rtype: int or list
        """
        is_many = isinstance(keys, list)
        if not is_many:
            keys = [keys]
        indexes = []
        if group == 'item':
            if not self._idmanager.itemid_mapped:
                self.build_itemid_map()
            indexes = [self._idmanager.itemid_map.get(k) for k in keys]
        elif group == 'user':
            if not self._idmanager.userid_mapped:
                self.build_userid_map()
            indexes = [self._idmanager.userid_map.get(k) for k in keys]
        if not is_many:
            return indexes[0]
        return np.array(indexes)

    # NOTE: Ugly naming?
    def get_index_pool(self, pool, group='item'):
        """Simple wrapper of get_index.
        For np.ndarray pool, it returns asis with nothing. But list, it perform get_index with keys in pool.

        :param pool: The list of keys.
        :param str group: Data group where to find (default: item)
        :rtype: np.ndarray
        """
        if isinstance(pool, list):
            pool = self.get_index(pool, group)
            pool = np.array([p for p in pool if p is not None])
        elif isinstance(pool, np.ndarray):
            pass
        else:
            raise ValueError('Unexpected type for pool: %s' % type(pool))
        assert isinstance(pool, np.ndarray)
        return pool


class Serializable(abc.ABC):
    def __init__(self, *args, **kwargs):
        pass

    def save(self, path=None, with_itemid_map=True, with_userid_map=True, data_fields=[]):
        if path is None:
            path = self.opt.model_path
        if with_itemid_map and not self._idmanager.itemid_mapped:
            self.build_itemid_map()
        if with_userid_map and not self._idmanager.userid_mapped:
            self.build_userid_map()
        data = self._get_data()
        if data_fields:
            data = [(k, v) for k, v in data if k in data_fields]
        with open(path, 'wb') as fout:
            total_objs = len(data)
            fout.write(struct.pack('Q', total_objs))
            for name, obj in data:
                bname = bytes(name, encoding='utf-8')
                fout.write(struct.pack('Q', len(bname)))
                fout.write(bname)
                s = pickle.dumps(obj, protocol=4)
                fout.write(struct.pack('Q', len(s)))
                fout.write(s)

    def _get_data(self):
        data = [('_idmanager', self._idmanager)]
        return data

    def load(self, path, data_fields=[]):
        with open(path, 'rb') as fin:
            total_objs = struct.unpack('Q', fin.read(8))[0]
            for _ in range(total_objs):
                name_sz = struct.unpack('Q', fin.read(8))[0]
                name = fin.read(name_sz).decode('utf8')
                obj_sz = struct.unpack('Q', fin.read(8))[0]
                if data_fields and name not in data_fields:
                    fin.seek(obj_sz, 1)
                    continue
                obj = pickle.loads(fin.read(obj_sz))
                setattr(self, name, obj)

    @classmethod
    def instantiate(cls, cls_opt, path, data_fields):
        opt = cls_opt().get_default_option()
        c = cls(opt)
        c.load(path, data_fields)
        return c


class TensorboardExtention(object):
    @abc.abstractmethod
    def get_evaluation_metrics(self):
        raise NotImplemented

    def _get_initial_tensorboard_data(self):
        tb = aux.Option({'summary_writer': None,
                         'name': None,
                         'metrics': {},
                         'feed_dict': {},
                         'merged_summary_op': None,
                         'session': None,
                         'pbar': None,
                         'data_root': None,
                         'step': 1})
        return tb

    def initialize_tensorboard(self, num_steps, name_prefix='', name_postfix='', metrics=None):
        if not self.opt.tensorboard:
            if not hasattr(self, '_tb_setted'):
                self.logger.debug('Cannot find tensorboard configuration.')
            self.tb_setted = False
            return
        name = self.opt.tensorboard.name
        name = name_prefix + name + name_postfix
        dtm = datetime.datetime.now().strftime('%Y%m%d-%H.%M')
        template = self.opt.tensorboard.get('name_template', '{name}.{dtm}')
        self._tb = self._get_initial_tensorboard_data()
        self._tb.name = template.format(name=name, dtm=dtm)
        if not os.path.isdir(self.opt.tensorboard.root):
            os.makedirs(self.opt.tensorboard.root)
        tb_dir = os.path.join(self.opt.tensorboard.root, self._tb.name)
        self._tb.data_root = tb_dir
        self._tb.summary_writer = tf.summary.FileWriter(tb_dir)
        if not metrics:
            metrics = self.get_evaluation_metrics()
        for m in metrics:
            self._tb.metrics[m] = tf.placeholder(tf.float32)
            tf.summary.scalar(m, self._tb.metrics[m])
            self._tb.feed_dict[self._tb.metrics[m]] = 0.0
        self._tb.merged_summary_op = tf.summary.merge_all()
        self._tb.session = tf.Session()
        self._tb.pbar = Progbar(num_steps, stateful_metrics=self._tb.metrics, verbose=0)
        self._tb_setted = True

    def update_tensorboard_data(self, metrics):
        if not self.opt.tensorboard:
            return
        metrics = [(m, np.float32(metrics.get(m, 0.0)))
                   for m in self._tb.metrics.keys()]
        self._tb.feed_dict = {self._tb.metrics[k]: v
                              for k, v in metrics}
        summary = self._tb.session.run(self._tb.merged_summary_op,
                                       feed_dict=self._tb.feed_dict)
        self._tb.summary_writer.add_summary(summary, self._tb.step)
        self._tb.pbar.update(self._tb.step, metrics)
        self._tb.step += 1

    def finalize_tensorboard(self):
        if not self.opt.tensorboard:
            return
        with open(os.path.join(self._tb.data_root, 'opt.json'), 'w') as fout:
            fout.write(json.dumps(self.opt, indent=2))
        self._tb.summary_writer.close()
        self._tb.session.close()
        self._tb = None
        tf.reset_default_graph()



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.string cimport string
from libc.stdint cimport int32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/cuda/als/als.hpp" namespace "cuda_als":
    cdef cppclass CuALS:
        CuALS() nogil except +
        bool init(string) nogil except +
        void set_placeholder(int64_t* lindptr, int64_t* rindptr, size_t batch_size);
        void initialize_model(float*, int,
                              float*, int) nogil except +
        void precompute(int) nogil except +
        pair[double, double] partial_update(int, int,
                             int64_t*, int32_t*, float*, int) nogil except +
        int get_vdim() nogil except +


cdef class CyALS:
    """CuALS object holder"""
    cdef CuALS* obj  # Cuda-ALS object

    def __cinit__(self):
        self.obj = new CuALS()

    def __dealloc__(self):
        del self.obj

    def init(self, opt_path):
        return self.obj.init(opt_path)

    def initialize_model(self,
                         np.ndarray[np.float32_t, ndim=2] P,
                         np.ndarray[np.float32_t, ndim=2] Q):
        self.obj.initialize_model(&P[0, 0], P.shape[0],
                                  &Q[0, 0], Q.shape[0])

    def set_placeholder(self, np.ndarray[np.int64_t, ndim=1] lindptr,
                        np.ndarray[np.int64_t, ndim=1] rindptr,
                        batch_size):
        self.obj.set_placeholder(&lindptr[0], &rindptr[0], batch_size)

    def precompute(self, axis):
        self.obj.precompute(axis)

    def get_vdim(self):
        return self.obj.get_vdim()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def partial_update(self,
                       int start_x,
                       int next_x,
                       np.ndarray[np.int64_t, ndim=1] indptr,
                       np.ndarray[np.int32_t, ndim=1] keys,
                       np.ndarray[np.float32_t, ndim=1] vals,
                       int axis):
        return self.obj.partial_update(start_x,
                                       next_x,
                                       &indptr[0],
                                       &keys[0],
                                       &vals[0],
                                       axis)



# -*- coding: utf-8 -*-
import numpy as np
import tensorflow as tf

from buffalo.misc import log


class TFALS(object):
    def __init__(self, opt, name="tf_als"):
        self.logger = log.get_logger("tf-als")
        self.opt = opt
        self.name = name
        self.sess = tf.Session()
        self.graph = tf.get_default_graph()

    def initialize_model(self, P, Q):
        with tf.variable_scope(self.name):
            self.P = tf.get_variable("P", initializer=P)
            self.Q = tf.get_variable("Q", initializer=Q)
            self.FF = tf.get_variable("FF", shape=(self.opt.d, self.opt.d),
                                      initializer=tf.keras.initializers.Zeros,
                                      dtype=tf.float32)
            self.build_graph()
        self.sess.run(tf.global_variables_initializer())

    def get_variable(self, name):
        return self.sess.run(getattr(self, name))

    def precompute(self, int_group):
        if int_group == 0:
            self.sess.run(self.precomputeQ)
        else:
            self.sess.run(self.precomputeP)

    def build_graph(self):
        self.start_x = tf.placeholder(dtype=tf.int32, name="start_x")
        self.next_x = tf.placeholder(dtype=tf.int32, name="next_x")
        self.rows = tf.placeholder(dtype=tf.int32, shape=(None, ), name="rows")
        self.keys = tf.placeholder(dtype=tf.int32, shape=(None, ), name="keys")
        self.vals = tf.placeholder(dtype=tf.float32, shape=(None, ), name="vals")
        for int_group in [0, 1]:
            self._build_graph(int_group)

    def _dot(self, X, Y):
        return tf.reduce_sum(X * Y, axis=1)

    def _build_graph(self, int_group):
        if int_group == 0:
            P, Q, reg = self.P, self.Q, self.opt.reg_u
        else:
            P, Q, reg = self.Q, self.P, self.opt.reg_i
        start_x, next_x, rows, keys, vals = \
            self.start_x, self.next_x, self.rows, self.keys, self.vals

        # compute ys
        Fgtr = tf.gather(Q, keys)
        coeff = self.vals * self.opt.alpha
        ys = tf.scatter_nd(tf.expand_dims(rows, axis=1),
                           Fgtr * tf.expand_dims(coeff + 1, axis=1),
                           shape=(next_x - start_x, self.opt.d))

        # prepare cg
        _P = P[start_x:next_x]
        Axs = tf.matmul(_P, self.FF) + reg * _P
        dots = self._dot(tf.gather(_P, rows), Fgtr)
        Axs = tf.tensor_scatter_add(Axs, tf.expand_dims(rows, axis=1),
                                    Fgtr * tf.expand_dims(dots * coeff, axis=1))
        rs = ys - Axs
        ps = rs
        rss_old = tf.reduce_sum(tf.square(rs), axis=1)

        # iterate cg steps
        for i in range(self.opt.num_cg_max_iters):
            Aps = tf.matmul(ps, self.FF) + ps * reg
            _dots = coeff * self._dot(tf.gather(ps, rows), Fgtr)
            Aps = tf.tensor_scatter_add(Aps, tf.expand_dims(rows, axis=1),
                                        Fgtr * tf.expand_dims(_dots, axis=1))
            pAps = self._dot(Aps, ps)
            alphas = rss_old / (pAps + self.opt.eps)
            _P = _P + ps * tf.expand_dims(alphas, axis=1)
            rs = rs - tf.expand_dims(alphas, axis=1) * Aps
            rss_new = tf.reduce_sum(tf.square(rs), axis=1)
            betas = rss_new / (rss_old + self.opt.eps)
            ps = rs + (tf.expand_dims(betas, axis=1) * ps)
            rss_old = rss_new

        if int_group == 1:
            if self.opt.compute_loss_on_training:
                self.err = tf.reduce_sum(tf.square(vals - dots))
            else:
                self.err = tf.constant(0.0, dtype=tf.float32)

        name = "updateP" if int_group == 0 else "updateQ"
        _update = P[start_x:next_x].assign(_P)
        with self.graph.control_dependencies([_update]):
            update = tf.constant(True)
        setattr(self, name, update)

        _FF = tf.assign(self.FF, tf.matmul(P, P, transpose_a=True))
        with self.graph.control_dependencies([_FF]):
            FF = tf.constant(True)
        name = "precomputeP" if int_group == 0 else "precomputeQ"
        setattr(self, name, FF)

    def _generate_rows(self, start_x, next_x, indptr):
        ends = indptr[start_x:next_x]
        begs = np.empty(next_x - start_x, dtype=np.int64)
        begs[0] = 0 if start_x == 0 else indptr[start_x - 1]
        begs[1:] = ends[:-1]
        ret = np.arange(next_x - start_x, dtype=np.int32)
        ret = np.repeat(ret, ends - begs)
        return ret, len(ret)

    def partial_update(self, start_x, next_x, indptr, keys, vals, int_group):
        rows, sz = self._generate_rows(start_x, next_x, indptr)
        feed_dict = {self.start_x: start_x, self.next_x: next_x,
                     self.rows: rows, self.keys: keys[:sz], self.vals: vals[:sz]}
        err = 0.0
        if int_group == 0:
            _ = self.sess.run(self.updateP, feed_dict=feed_dict)
        else:
            _, _err = self.sess.run([self.updateQ, self.err], feed_dict=feed_dict)
            err += _err
        return err



# -*- coding: utf-8 -*-
import os
import psutil
import traceback

import h5py
import numpy as np

from buffalo.data import prepro
from buffalo.misc import aux, log
from buffalo.data.base import Data, DataOption


class MatrixMarketOptions(DataOption):
    def get_default_option(self) -> aux.Option:
        opt = {
            'type': 'matrix_market',
            'input': {
                'main': '',
                'uid': '',  # if not set, row-id is used as userid.
                'iid': ''  # if not set, col-id is used as itemid.
            },
            'data': {
                'internal_data_type': 'matrix',
                'validation': {
                    'name': 'sample',
                    'p': 0.01,
                    'max_samples': 500
                },
                'batch_mb': 1024,
                'use_cache': False,
                'tmp_dir': '/tmp/',
                'path': './mm.h5py'
            }
        }
        return aux.Option(opt)

    def is_valid_option(self, opt) -> bool:
        assert super(MatrixMarketOptions, self).is_valid_option(opt)
        if not opt['type'] == 'matrix_market':
            raise RuntimeError('Invalid data type: %s' % opt['type'])
        if opt['data']['internal_data_type'] != 'matrix':
            raise RuntimeError('MatrixMarket only support internal data type(matrix)')
        return True


class MatrixMarket(Data):
    def __init__(self, opt, *args, **kwargs):
        super().__init__(opt, *args, **kwargs)
        self.name = 'MatrixMarket'
        self.logger = log.get_logger('MatrixMarket')
        if isinstance(self.value_prepro,
                      (prepro.SPPMI)):
            raise RuntimeError(f'{self.opt.data.value_prepro.name} does not support MatrixMarket')
        self.data_type = 'matrix'

    def _create(self, data_path, P, H):
        def get_max_column_length(fname):
            with open(fname) as fin:
                max_col = 0
                for l in fin:
                    max_col = max(max_col, len(l))
            return max_col

        uid_path, iid_path, main_path = P['uid_path'], P['iid_path'], P['main_path']
        num_users, num_items, num_nnz = map(int, H.split())
        # Manually updating progress bar is a bit naive
        with log.pbar(log.DEBUG, total=5, mininterval=30) as pbar:
            uid_max_col = len(str(num_users)) + 1
            if uid_path:
                uid_max_col = get_max_column_length(uid_path) + 1
            pbar.update(1)
            iid_max_col = len(str(num_items)) + 1
            if iid_path:
                iid_max_col = get_max_column_length(iid_path) + 1
            pbar.update(1)
            try:
                db = self._create_database(data_path,
                                           num_users=num_users,
                                           num_items=num_items,
                                           num_nnz=num_nnz,
                                           uid_max_col=uid_max_col,
                                           iid_max_col=iid_max_col)
                idmap = db['idmap']
                # if not given, assume id as is
                if uid_path:
                    with open(uid_path) as fin:
                        idmap['rows'][:] = np.loadtxt(fin, dtype=f'S{uid_max_col}')
                else:
                    idmap['rows'][:] = np.array([str(i) for i in range(1, num_users + 1)],
                                                dtype=f'S{uid_max_col}')
                pbar.update(1)
                if iid_path:
                    with open(iid_path) as fin:
                        idmap['cols'][:] = np.loadtxt(fin, dtype=f'S{iid_max_col}')
                else:
                    idmap['cols'][:] = np.array([str(i) for i in range(1, num_items + 1)],
                                                dtype=f'S{iid_max_col}')
                pbar.update(1)
                num_header_lines = 0
                with open(main_path) as fin:
                    for line in fin:
                        if line.strip().startswith('%'):
                            num_header_lines += 1
                        else:
                            break
                pbar.update(1)
            except Exception as e:
                self.logger.error('Cannot create db: %s' % (str(e)))
                self.logger.error(traceback.format_exc())
                raise
        return db, num_header_lines

    def _create_working_data(self, db, source_path, ignore_lines):
        """
        Args:
            source_path: source data file path
            ignore_lines: number of lines to skip from start line
        """
        vali_indexes = [] if 'vali' not in db else db['vali']['indexes']
        vali_lines = []
        file_path = aux.get_temporary_file(self.opt.data.tmp_dir)
        with open(file_path, 'w') as w:
            fin = open(source_path, mode='r')
            file_size = fin.seek(0, 2)
            fin.seek(0, 0)
            for _ in range(ignore_lines):
                fin.readline()
            total = file_size - fin.tell()
            buffered = ''
            CHUNK_SIZE = 4096 * 1000
            total_lines = 0
            vali_indexes = sorted(vali_indexes)
            target_index = vali_indexes[0] if vali_indexes else -1
            vali_indexes = vali_indexes[1:]
            with log.pbar(log.INFO, total=total, mininterval=10) as pbar:
                while True:
                    buffered += fin.read(CHUNK_SIZE)
                    if buffered == '':
                        break
                    current_file_position = fin.tell()
                    pbar.update(CHUNK_SIZE)
                    num_lines_on_buffer = buffered.count('\n')
                    # search the position of validation sample and extract
                    # it from training data
                    while target_index >= 0 and target_index <= (total_lines + num_lines_on_buffer):
                        no_line = total_lines
                        new_buffered = ''
                        from_index = 0
                        for idx, c in enumerate(buffered):
                            if c == '\n':
                                if no_line == target_index:
                                    vali_lines.append(buffered[from_index:idx])
                                    if from_index > 0:
                                        w.write(buffered[0:from_index])
                                    new_buffered = buffered[idx + 1:]
                                    no_line += 1
                                    total_lines += 1
                                    num_lines_on_buffer -= 1
                                    break
                                no_line += 1
                                total_lines += 1
                                from_index = idx + 1
                                num_lines_on_buffer -= 1
                        buffered = new_buffered
                        if vali_indexes:
                            target_index, vali_indexes = vali_indexes[0], vali_indexes[1:]
                        else:
                            target_index = -1
                    where = buffered.rfind('\n')
                    total_lines += num_lines_on_buffer
                    if where != -1:
                        w.write(buffered[:where + 1])
                        buffered = buffered[where + 1:]
                    elif current_file_position == file_size:
                        w.write(buffered)
                        buffered = ''
            w.close()
            fin.close()
            return w.name, vali_lines

    def create(self) -> h5py.File:
        mm_main_path = self.opt.input.main
        mm_uid_path = self.opt.input.uid
        mm_iid_path = self.opt.input.iid

        data_path = self.opt.data.path
        if os.path.isfile(data_path) and self.opt.data.use_cache:
            self.logger.info('Use cached DB on %s' % data_path)
            self.open(data_path)
            return

        self.logger.info('Create the database from matrix market file.')
        with open(mm_main_path) as fin:
            header = '%'
            while header.startswith('%'):
                header = fin.readline()
            self.logger.debug('Building meta part...')
            db, num_header_lines = self._create(data_path,
                                                {'main_path': mm_main_path,
                                                 'uid_path': mm_uid_path,
                                                 'iid_path': mm_iid_path},
                                                header)
            try:
                num_header_lines += 1  # add metaline
                self.logger.info('Creating working data...')
                tmp_main, validation_data = self._create_working_data(db,
                                                                      mm_main_path,
                                                                      num_header_lines)
                self.logger.debug(f'Working data is created on {tmp_main}')
                self.logger.info('Building data part...')
                self._build_data(db, tmp_main, validation_data)
                db.attrs['completed'] = 1
                db.close()
                self.handle = h5py.File(data_path, 'r')
                self.path = data_path
            except Exception as e:
                self.logger.error('Cannot create db: %s' % (str(e)))
                self.logger.error(traceback.format_exc().splitlines())
                raise
            finally:
                if hasattr(self, 'patr'):
                    if os.path.isfile(self.path):
                        os.remove(self.path)
        self.logger.info('DB built on %s' % data_path)



# -*- coding: utf-8 -*-
from buffalo.misc import aux


def load(opt):
    from buffalo.data.mm import MatrixMarket
    from buffalo.data.stream import Stream
    if isinstance(opt, (str,)):
        opt = aux.Option(opt)

    assert isinstance(opt, (dict, aux.Option)), 'opt must be either str, or dict/aux.Option but {}'.format(type(opt))
    if opt['type'] == 'matrix_market':
        return MatrixMarket(opt)
    if opt['type'] == 'stream':
        return Stream(opt)
    raise RuntimeError('Unexpected data.type: {}'.format(opt['type']))



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.string cimport string
from libc.stdint cimport int64_t, int32_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/algo_impl/als/als.hpp" namespace "als":
    cdef cppclass CALS:
        void release() nogil except +
        bool init(string) nogil except +
        void initialize_model(float*, int,
                              float*, int) nogil except +
        void precompute(int) nogil except +
        pair[double, double] partial_update(int,
                                            int,
                                            int64_t*,
                                            int32_t*,
                                            float*,
                                            int) nogil except +



cdef class CyALS:
    """CALS object holder"""
    cdef CALS* obj  # C-ALS object

    def __cinit__(self):
        self.obj = new CALS()

    def __dealloc__(self):
        self.obj.release()
        del self.obj

    def init(self, option_path):
        return self.obj.init(option_path)

    def initialize_model(self,
                         np.ndarray[np.float32_t, ndim=2] P,
                         np.ndarray[np.float32_t, ndim=2] Q):
        self.obj.initialize_model(&P[0, 0], P.shape[0],
                                  &Q[0, 0], Q.shape[0])

    def precompute(self, axis):
        self.obj.precompute(axis)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def partial_update(self,
                       int start_x,
                       int next_x,
                       np.ndarray[np.int64_t, ndim=1] indptr,
                       np.ndarray[np.int32_t, ndim=1] keys,
                       np.ndarray[np.float32_t, ndim=1] vals,
                       int axis):
        return self.obj.partial_update(start_x,
                                       next_x,
                                       &indptr[0],
                                       &keys[0],
                                       &vals[0],
                                       axis)



#!/usr/bin/env python
# Converting fvec-files from the texmex collection:
# http://corpus-texmex.irisa.fr/
# Should be run on an Intel (little-endian) machine

import sys
import struct
import array
import os

ifile = sys.argv[1]
maxQty = 0
if len(sys.argv) == 3:
  maxQty = int(sys.argv[2])

fsize = os.path.getsize(ifile)

fin = open(ifile, 'rb')

dim = struct.unpack('i', fin.read(4))[0]

if fsize % (4*(dim + 1)) != 0:
  raise Exception("File: " + ifile + " has the wrong format, expected dim = " + str(dim))

fin.close()


fin = open(ifile, 'rb')

rowQty = fsize / (dim + 1) / 4

#print rowQty, dim

if maxQty != 0:
  rowQty = min(rowQty, maxQty)

for i in range(0,int(rowQty)):
  tmpd = struct.unpack('<i', fin.read(4))[0]
  if tmpd != dim:
    raise Exception("Wrong format, got wrong dimension: " + str(tmpd) + " expect: " + str(dim) + " row " + str(i+1))

  vec = array.array('f')
  
  vec.fromfile(fin, dim)
  out = ''
  for j in range(0,dim):
    if j > 0: out = out + ' '
    out = out + str(vec[j])
  print(out)
    



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp cimport bool
from libcpp.string cimport string
from libc.stdint cimport int32_t, uint32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/algo_impl/w2v/w2v.hpp" namespace "w2v":
    cdef cppclass CW2V:
        void release() nogil except +
        bool init(string) nogil except +
        void initialize_model(float*, int32_t,
                              int32_t*,
                              uint32_t*,
                              int32_t*,
                              int64_t) nogil except +
        void launch_workers()
        void add_jobs(int,
                      int,
                      int64_t*,
                      int32_t*) nogil except +
        double join()


cdef class CyW2V:
    """CW2V object holder"""
    cdef CW2V* obj  # C-W2V object

    def __cinit__(self):
        self.obj = new CW2V()

    def __dealloc__(self):
        self.obj.release()
        del self.obj

    def init(self, option_path):
        return self.obj.init(option_path)

    def initialize_model(self,
                         np.ndarray[np.float32_t, ndim=2] L0,
                         np.ndarray[np.int32_t, ndim=1] index,
                         np.ndarray[np.uint32_t, ndim=1] scale,
                         np.ndarray[np.int32_t, ndim=1] dist,
                         int64_t total_word_count):
        self.obj.initialize_model(&L0[0, 0],
                                  L0.shape[0],
                                  &index[0],
                                  &scale[0],
                                  &dist[0],
                                  total_word_count)

    def launch_workers(self):
        self.obj.launch_workers()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_jobs(self,
                 int start_x,
                 int next_x,
                 np.ndarray[np.int64_t, ndim=1] indptr,
                 np.ndarray[np.int32_t, ndim=1] sequences):
        self.obj.add_jobs(start_x,
                          next_x,
                          &indptr[0],
                          &sequences[0])


    def join(self):
        return self.obj.join()



# -*- coding: utf-8 -*-
from buffalo.misc.aux import InputOptions, Option


class AlgoOption(InputOptions):
    def __init__(self, *args, **kwargs):
        super(AlgoOption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """Default options for Algo classes.

        :ivar bool evaluation_on_learning: Set True to do run evaluation on training phrase. (default: True)
        :ivar bool compute_loss_on_training: Set True to calculate loss on training phrase. (default: True)
        :ivar int early_stopping_rounds: The number of exceed epochs after reached minimum loss on training phrase. If set 0, it doesn't work. (default: 0)
        :ivar bool save_best: Whenver the loss improved, save the model.
        :ivar int evaluation_period: How often will do evaluation in epochs. (default: 1)
        :ivar int save_period: How often will do save_best routine in epochs. (default: 10)
        :ivar int random_seed: Random Seed
        :ivar dict validation: The validation options.
        """
        opt = {
            'evaluation_on_learning': True,
            'compute_loss_on_training': True,
            'early_stopping_rounds': 0,
            'save_best': False,
            'evaluation_period': 1,
            'save_period': 10,
            'random_seed': 0,
            'validation': {}
        }
        return opt

    def get_default_optimize_option(self):
        """Default options for optimize feature.

        :ivar str loss: Name of loss to monitor. (default: train_loss)
        """
        opt = {
            'loss': 'train_loss',
        }
        return opt

    def get_default_tensorboard_option(self):
        """Default options for tensorboard feature.

        :ivar str name: Name of graph name. (default: default)
        :ivar str root: Path where to make graph directory. (default: ./tb)
        :ivar str name_template: Name template for directory name. (default: {name}.{dtm})
        """
        opt = {
            'name': 'default',
            'root': './tb',
            'name_template': '{name}.{dtm}'
        }
        return opt

    def is_valid_option(self, opt):
        b = super().is_valid_option(opt)
        for f in ['num_workers']:
            if not f in opt:
                raise RuntimeError(f'{f} not defined')
        return b


class ALSOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super(ALSOption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Alternating Least Squares.

        :ivar bool adaptive_reg: Set True, for adaptive regularization. (default: False)
        :ivar bool save_factors: Set True, to save models. (default: False)
        :ivar bool accelerator: Set True, to accelerate training using GPU. (default: False)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int num_iters: The number of iterations for training. (default: 10)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int hyper_threads: The number of hyper threads when using cuda cores. (default: 256)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.1)
        :ivar float reg_i: The L2 regularization coefficient for item embedding matrix. (default: 0.1)
        :ivar float alpha: The coefficient of giving more weights to losses on positive samples. (default: 8)
        :ivar float eps: epsilon for numerical stability (default: 1e-10)
        :ivar float cg_tolerance: tolerance of conjugate gradient for early stopping iterations (default: 1e-10)
        :ivar str optimizer: The name of optimizer, should be in [llt, ldlt, manual_cg, eigen_cg, eigen_bicg, eigen_gmres, eigen_dgmres, eigen_minres]. (default: manual_cg)
        :ivar int num_cg_max_iters: The number of maximum iterations for conjuaget gradient optimizer. (default: 3)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """

        opt = super().get_default_option()
        opt.update({
            'adaptive_reg': False,
            'save_factors': False,
            'accelerator': False,

            'd': 20,
            'num_iters': 10,
            'num_workers': 1,
            'hyper_threads': 256,
            'num_cg_max_iters': 3,
            'reg_u': 0.1,
            'reg_i': 0.1,
            'alpha': 8,
            'optimizer': 'manual_cg',
            'cg_tolerance': 1e-10,
            'eps': 1e-10,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization Options for ALS.

        :ivar str loss: Target loss to optimize.
        :ivar int max_trials: The maximum experiments for optimization. If not given, run forever.
        :ivar int min_trials: The minimum experiments before deploying model. (Since the best parameter may not be found after `min_trials`, the first best parameter is always deployed)
        :ivar bool deployment: Set True to train model with the best parameter. During the optimization, it try to dump the model which beated the previous best loss.
        :ivar bool start_with_default_parameters: If set to True, the loss value of the default parameter is used as the starting loss to beat.
        :ivar dict space: The parameter space definition. For more information, pleases reference hyperopt's express. Note) Due to hyperopt's `randint` does not provide lower value, we had to implement it a bait tricky. Pleases see optimize.py to check how we deal with `randint`.k
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'adaptive_reg': ['choice', ['adaptive_reg', [0, 1]]],
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]],
                'alpha': ['randint', ['alpha', 1, 32]]
            }
        })
        return Option(opt)

class CFROption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super(CFROption, self).__init__(*args, **kwargs)

    def get_default_option(self):
        """ Basic Options for CoFactor.

        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int num_iters: The number of iterations for training. (default: 10)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.1)
        :ivar float reg_i: The L2 regularization coefficient for item embedding matrix. (default: 0.1)
        :ivar float reg_c: The L2 regularization coefficient for context embedding matrix. (default: 0.1)
        :ivar float eps: epsilon for numerical stability (default: 1e-10)
        :ivar float cg_tolerance: The tolerance for early stopping conjugate gradient optimizer. (default: 1e-10)
        :ivar float alpha: The coefficient of giving more weights to losses on positive samples. (default: 8.0)
        :ivar float l: The relative weight of loss on user-item relation over item-context relation. (default: 1.0)
        :ivar str optimizer: The name of optimizer, should be in [llt, ldlt, manual_cg, eigen_cg, eigen_bicg, eigen_gmres, eigen_dgmres, eigen_minres]. (default: manual_cg)
        :ivar int num_cg_max_iters: The number of maximum iterations for conjuaget gradient optimizer. (default: 3)
        :ivar str model_path: Where to save model. (default: '')
        :ivar dict data_opt: This options will be used to load data if given. (default: {})
        """
        opt = super().get_default_option()
        opt.update({
            'save_factors': False,
            'd': 20,
            'num_iters': 10,
            'num_workers': 1,
            'num_cg_max_iters': 3,

            'cg_tolerance': 1e-10,
            'eps': 1e-10,
            'reg_u': 0.1,
            'reg_i': 0.1,
            'reg_c': 0.1,
            'alpha': 8.0,
            'l': 1.0,

            'optimizer': 'manual_cg',
            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for CoFactor.

        :ivar str loss: Target loss to optimize.
        :ivar int max_trials: Maximum experiments for optimization. If not given, run forever.
        :ivar int min_trials: Minimum experiments before deploying model. (Since the best parameter may not be found after `min_trials`, the first best parameter is always deployed)
        :ivar bool deployment(: Set True to train model with the best parameter. During the optimization, it try to dump the model which beated the previous best loss.
        :ivar bool start_with_default_parameters: If set to True, the loss value of the default parameter is used as the starting loss to beat.
        :ivar dict space: Parameter space definition. For more information, pleases reference hyperopt's express. Note) Due to hyperopt's `randint` does not provide lower value, we had to implement it a bait tricky. Pleases see optimize.py to check how we deal with `randint`.k
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]],
                'reg_c': ['uniform', ['reg_i', 0.1, 1]],
                'alpha': ['randint', ['alpha', 1, 32]],
                'l': ['randint', ['alpha', 1, 32]]
            }
        })
        return Option(opt)

    def is_valid_option(self, opt):
        b = super().is_valid_option(opt)
        possible_optimizers = ["llt", "ldlt", "manual_cg", "eigen_cg", "eigen_bicg",
                               "eigen_gmres", "eigen_dgmres", "eigen_minres"]
        if not opt.optimizer in possible_optimizers:
            msg = f"optimizer ({opt.optimizer}) should be in {possible_optimizers}"
            raise RuntimeError(msg)
        return b


class BPRMFOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Bayesian Personalized Ranking Matrix Factorization.

        :ivar bool use_bias: Set True, to use bias term for the model.
        :ivar int evaluation_period: (default: 100)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int num_iters: The number of iterations for training. (default: 100)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar bool update_i: Set True, to update positive item feature. (default: True)
        :ivar bool update_j: Set True, to update negative item feature. (default: True)
        :ivar float reg_u: The L2 regularization coefficient for user embedding matrix. (default: 0.025)
        :ivar float reg_i: The L2 regularization coefficient for positive item embedding matrix. (default: 0.025)
        :ivar float reg_j: The L2 regularization coefficient for negative item embedding matrix. (default: 0.025)
        :ivar float reg_b: The L2 regularization coefficient for bias term. (default: 0.025)
        :ivar str optimizer: The name of optimizer, should be one of [sgd, adam]. (default: sgd)
        :ivar float lr: The learning rate.
        :ivar float min_lr: The minimum of learning rate, to prevent going to zero by learning rate decaying. (default: 0.0001)
        :ivar float beta1: The parameter of Adam optimizer. (default: 0.9)
        :ivar float beta2: The parameter of Adam optimizer. (default: 0.999)
        :ivar bool per_coordinate_normalize: This is a bit tricky option for Adam optimizer. Before update factors with graidents, do normalize gradients per class by its number of contributed samples. (default: False)
        :ivar int num_negative_samples: The number of negaitve samples. (default: 1)
        :ivar float sampleling_power: This paramemter control sampling distribution. When it set to 0, it draw negative items from uniform distribution, while to set 1, it draw from the given data popularation. (default: 0.0)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """
        opt = super().get_default_option()
        opt.update({
            'use_bias': True,
            'evaluation_period': 100,
            'num_workers': 1,
            'num_iters': 100,
            'd': 20,
            'update_i': True,
            'update_j': True,
            'reg_u': 0.025,
            'reg_i': 0.025,
            'reg_j': 0.025,
            'reg_b': 0.025,

            'optimizer': 'sgd',
            'lr': 0.002,
            'min_lr': 0.0001,
            'beta1': 0.9,
            'beta2': 0.999,

            'per_coordinate_normalize': False,
            'num_negative_samples': 1,
            'sampling_power': 0.0,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for BPRMF.
        """
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'reg_u': ['uniform', ['reg_u', 0.1, 1]],
                'reg_i': ['uniform', ['reg_i', 0.1, 1]]
            }
        })
        return Option(opt)


class W2VOption(AlgoOption):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_default_option(self):
        """Options for Word2Vec.

        :ivar bool evaluation_on_learning: Set True to do run evaluation on training phrase. (default: False)
        :ivar int num_workers: The number of threads. (default: 1)
        :ivar int num_iters: The number of iterations for training. (default: 100)
        :ivar int d: The number of latent feature dimension. (default: 20)
        :ivar int window: The window size. (default: 5)
        :ivar int min_count: The minimum required frequency of the words to use training vocabulary. (default: 5)
        :ivar float sample: The sampling ratio to downsample the frequent words. (default: 0.001)
        :ivar float lr: The learning rate.
        :ivar int num_negative_samples: The number of negaitve samples. (default: 5)
        :ivar str model_path: Where to save model.
        :ivar dict data_opt: This options will be used to load data if given.
        """
        opt = super().get_default_option()
        opt.update({
            'evaluation_on_learning': False,

            'num_workers': 1,
            'num_iters': 3,
            'd': 20,
            'window': 5,
            'min_count': 5,
            'sample': 0.001,

            'lr': 0.025,
            'min_lr': 0.0001,

            'num_negative_samples': 5,

            'model_path': '',
            'data_opt': {}
        })
        return Option(opt)

    def get_default_optimize_option(self):
        """Optimization options for W2V
        """
        #
        opt = super().get_default_optimize_option()
        opt.update({
            'loss': 'train_loss',
            'max_trials': 100,
            'min_trials': 0,
            'deployment': True,
            'start_with_default_parameters': True,
            'space': {
                'd': ['randint', ['d', 10, 30]],
                'window': ['randint', ['window', 2, 8]],
                'num_negative_samples': ['randint', ['alpha', 1, 12]]
            }
        })
        return Option(opt)



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool as bool_t
from libc.stdint cimport int32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/algo_impl/cfr/cfr.hpp" namespace "cfr":
    cdef cppclass CCFR:
        bool_t init(string) nogil except +
        void set_embedding(float*, int, string) nogil except +
        void precompute(string) nogil except +
        double partial_update_user(int, int,
                                   int64_t*, int32_t*, float*) nogil except +
        double partial_update_item(int, int,
                                   int64_t*, int32_t*, float*,
                                   int64_t*, int32_t*, float*) nogil except +
        double partial_update_context(int, int,
                                      int64_t*, int32_t*, float*) nogil except +


cdef class CyCFR:
    """CCFR object holder"""
    cdef CCFR* obj  # C-CFR object

    def __cinit__(self):
        self.obj = new CCFR()

    def __dealloc__(self):
        del self.obj

    def init(self, opt_path):
        return self.obj.init(opt_path)

    def precompute(self, obj_type):
        self.obj.precompute(obj_type)

    def set_embedding(self,
                      np.ndarray[np.float32_t, ndim=2] F, obj_type):
        self.obj.set_embedding(&F[0, 0], F.shape[0], obj_type)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def partial_update_user(self, int start_x, int next_x,
                            np.ndarray[np.int64_t, ndim=1] indptrs,
                            np.ndarray[np.int32_t, ndim=1] keys,
                            np.ndarray[np.float32_t, ndim=1] vals):
        return self.obj.partial_update_user(start_x, next_x,
                                            &indptrs[0], &keys[0], &vals[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def partial_update_item(self, int start_x, int next_x,
                            np.ndarray[np.int64_t, ndim=1] indptrs_u,
                            np.ndarray[np.int32_t, ndim=1] keys_u,
                            np.ndarray[np.float32_t, ndim=1] vals_u,
                            np.ndarray[np.int64_t, ndim=1] indptrs_c,
                            np.ndarray[np.int32_t, ndim=1] keys_c,
                            np.ndarray[np.float32_t, ndim=1] vals_c):
        return self.obj.partial_update_item(start_x, next_x,
                                            &indptrs_u[0], &keys_u[0], &vals_u[0],
                                            &indptrs_c[0], &keys_c[0], &vals_c[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def partial_update_context(self, int start_x, int next_x,
                               np.ndarray[np.int64_t, ndim=1] indptrs,
                               np.ndarray[np.int32_t, ndim=1] keys,
                               np.ndarray[np.float32_t, ndim=1] vals):
        return self.obj.partial_update_context(start_x, next_x,
                                               &indptrs[0], &keys[0], &vals[0])
